from __future__ import annotations
from urllib.request import HTTPError
from ._base import _BaseKube
from arcgis.gis import GIS
from arcgis.gis._impl._con import Connection
from arcgis._impl.common._mixins import PropertyMap
from typing import Dict, Any, List, Tuple, Optional


###########################################################################
class Overview(_BaseKube):
    """ """

    _con = None
    _gis = None
    _url = None

    # ---------------------------------------------------------------------
    def __init__(self, url: str, gis: GIS) -> None:
        """initializer"""
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    # ----------------------------------------------------------------------
    def _init(self, connection: Optional[Connection] = None) -> None:
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f": "json", "resource": "all"}
        try:
            result = connection.get(path=self._url, params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})

    # ---------------------------------------------------------------------
    def get(self, resource: Optional[str] = None) -> Dict[str, Any]:
        """
        The `get` returns the persisted cache or real-time
        information, such as health or status information, for the overview
        resource types, and is what is called when the Overview page of the
        Enterprise Manager is updated. Whether the information is cached or
        returned in real-time depends on the updateIntervalMin property
        returned by the config resource, which specifies the interval (in
        minutes) from which information for each resource type is pulled
        and cached.

        Resource types that have an updateIntervalMin value of 0 will not
        have their information cached and, instead, will have real-time
        information returned when the resource is called. The update
        interval can me modified through the update operation.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        resource            Optional String.  Specifies the resource type (criticalLogs,
                            dataStores, systemServices, utilityServices) that will have their
                            information returned. Using all as the input value will return
                            information for all resource types.

                            The default is `None`. When `None`, all resources will be returned.
        ===============     ====================================================================

        :return: Dict[str, Any]

        """
        if resource is None:
            resource = "all"
        url = self._url
        params = {"f": "json", "resources": resource}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @property
    def config(self) -> Dict[str, Any]:
        """
        Gets/sets a dictionary of resource types that correspond with the
        `Overview` class.  It contains the update interval for each
        property.


        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        resource            Required Dictionary. A dictionary object containing the `id`, `type`,
                            and `updateIntervalMin` for an overview resource type, returned by
                            the config resource. The accepted values for `updateIntervalMin`
                            (0-60) can be modified to update the interval (in minutes) of which
                            the resource type will have it's information pulled and cached. If
                            set to 0, information for the resource type will not be cached and
                            will, instead, have it's real-time information returned when the
                            overview resource is called. The default values for each resource
                            type are listed below:

                               + criticalLogs: 0
                               + systemServices: 1
                               + utilityServices: 1
                               + dataStores: 2

        ===============     ====================================================================

        :return: Dict[str, Any]

        """
        url = f"{self._url}/config"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @config.setter
    def config(self, value: Dict[str, Any]) -> None:
        """
        See main ``config`` property docstring
        """
        url = f"{self._url}/config/config"
        params = {"f": "json", "resourceConfigJson": value}
        return self._con.get(url, params)



from ._codetemplate import imagets_classifier_prf
import json
import traceback
import numpy as np
import pandas as pd
from IPython.display import display

from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel, _EmptyData

try:
    from ._psetae_utils import FocalLoss, miou
    from ._psetae_utils import PseTae, weight_init, model_eval
    from .._data_utils.psetae_data import show_results
    from .._utils.common import _get_emd_path
    from pathlib import Path
    from fastai.vision import Learner, partial, optim

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class PSETAE(ArcGISModel):

    """
    Creates a Pixel-Set encoder + Temporal Attention Encoder sequence classifier.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    **Keyword Arguments**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    mlp1                    Optional list. Dimensions of the
                            successive feature spaces of MLP1.
                            default set to [32, 64]
    ---------------------   -------------------------------------------
    pooling                 Optional string. Pixel-embedding
                            pooling strategy, can be chosen in
                            ('mean','std','max','min').
                            default set to 'mean'
    ---------------------   -------------------------------------------
    mlp2                    Optional list. Dimensions of the
                            successive feature spaces of MLP2.
                            default set to [128, 128]
    ---------------------   -------------------------------------------
    n_head                  Optional integer. Number of attention heads.
                            default set to 4
    ---------------------   -------------------------------------------
    d_k                     Optional integer. Dimension of the
                            key and query vectors. default set to 32
    ---------------------   -------------------------------------------
    dropout                 Optional float. dropout. default set to 0.2
    ---------------------   -------------------------------------------
    T                       Optional integer. Period to use for
                            the positional encoding.
                            default set to 1000
    ---------------------   -------------------------------------------
    mlp4                    Optional list. dimensions of decoder mlp
                            .default set to [64, 32]
    =====================   ===========================================

    :return: `PSETAE` Object
    """

    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data, pretrained_path=None, *args, **kwargs)

        self.kwargs = kwargs
        gamma = kwargs.get("gamma", 1)
        psetae = PseTae(
            input_dim=data._n_channel,
            len_max_seq=data._n_temp,
            positions=data._date_positions,
            out_class=len(data._class_map_dict),
            **kwargs,
        )

        psetae.apply(weight_init)

        self.learn = Learner(
            data,
            psetae,
            loss_func=FocalLoss(gamma),
            opt_func=partial(optim.Adam, betas=(0.5, 0.99)),
            metrics=[miou(len(data._class_map_dict))],
        )

        self.learn.model = self.learn.model.to(self._device)
        self.learn.model._device = self._device
        self._slice_lr = False
        if pretrained_path is not None:
            self.load(pretrained_path)
        self._code = imagets_classifier_prf

        def __str__(self):
            return self.__repr__()

        def __repr__(self):
            return "<%s>" % (type(self).__name__)

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_psetae_inferencing"
        _emd_template["Kwargs"] = self.kwargs
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageTsClassifier.py"
        else:
            _emd_template[
                "InferenceFunction"
            ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageTsClassifier.py"
        _emd_template["ModelType"] = "ImageClassification"
        _emd_template["Class_mapping"] = self._data._class_map_dict
        if self._data._num_class_map_dict:
            _emd_template["Num_class_mapping"] = self._data._num_class_map_dict
        _emd_template["n_channel"] = self._data._n_channel
        _emd_template["n_temporal"] = self._data._n_temp
        _emd_template["IsMultidimensional"] = self._data._is_multidimensional
        _emd_template["date_positions"] = self._data._date_positions
        _emd_template["ImageHeight"] = 256
        _emd_template["ImageWidth"] = 256
        _emd_template["ImageSpaceUsed"] = self._data._imagespace
        _emd_template["convertmap"] = (
            self._data._convertmap if self._data._convertmap else None
        )
        _emd_template["bandindex"] = (
            self._data._bandindex if self._data._bandindex else None
        )
        _emd_template["timeindex"] = (
            self._data._timeindex if self._data._timeindex else None
        )
        _emd_template["timestep_infer"] = self._data._timestep_infer
        _emd_template["channels_infer"] = self._data._channels_infer
        _emd_template["mean_norm_stats"] = {
            "mean_stats": [
                list(i) for i in (self._data._mean_norm_stats).astype(np.float64)
            ]
        }
        _emd_template["std_norm_stats"] = {
            "std_stats": [
                list(i) for i in (self._data._std_norm_stats).astype(np.float64)
            ]
        }

        return _emd_template

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a PSETAE object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return: `PSETAE` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        chip_size = emd["ImageHeight"]
        kwargs = emd.get("Kwargs", {})

        if "backbone" in kwargs.keys():
            kwargs.pop("backbone")

        if data is None:
            data = _EmptyData(
                path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
            )
            data._n_channel = emd.get("n_channel", None)
            data._n_temp = emd.get("n_temporal", None)
            data._mean_norm_stats = emd.get("mean_norm_stats", None)
            data._std_norm_stats = emd.get("std_norm_stats", None)
            data._class_map_dict = emd.get("Class_mapping", None)
            data._date_positions = emd.get("date_positions", None)
            data._convertmap = emd.get("convertmap", None)
            data.emd_path = emd_path
            data.emd = emd
            data._is_empty = True
        return cls(data, **model_params, pretrained_path=str(model_file), **kwargs)

    @property
    def _model_metrics(self):
        return self.compute_metrics()

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return PSETAE._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["RCNN_Masks"]

    def show_results(self, rows=20, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================
        total_sample_size       Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================
        **kwargs**

        """
        show_results(self, rows, **kwargs)

    def compute_metrics(self):
        """
        Computes mean intersection over union (mIOU) and
        overall accuracy (OA) on validation set.

        """
        if not hasattr(self._data, "load_empty"):
            raise Exception("Dataset is required for compute metrics")
        class_dict = (
            self._data._num_class_map_dict
            if self._data._num_class_map_dict
            else self._data._class_map_dict
        )
        mats, miou = model_eval(
            self._data, self.learn.model, class_dict, self._data._convertmap
        )
        return {
            "mIOU": "{}".format(miou),
            "Accuracy (OA)": "{}".format(mats[1]["Accuracy"]),
        }

    def accuracy(self):
        """
        Computes overall accuracy (OA) on validation set.

        """
        if not hasattr(self._data, "load_empty"):
            raise Exception("Dataset is required for compute metrics")
        class_dict = (
            self._data._num_class_map_dict
            if self._data._num_class_map_dict
            else self._data._class_map_dict
        )
        mats, miou = model_eval(
            self._data, self.learn.model, class_dict, self._data._convertmap
        )
        return {"Accuracy (OA)": "{}".format(mats[1]["Accuracy"])}

    def mIOU(self):
        """
        Computes mean intersection over union (mIOU) on validation set.

        """
        if not hasattr(self._data, "load_empty"):
            raise Exception("Dataset is required for compute metrics")
        class_dict = (
            self._data._num_class_map_dict
            if self._data._num_class_map_dict
            else self._data._class_map_dict
        )
        mats, miou = model_eval(
            self._data, self.learn.model, class_dict, self._data._convertmap
        )
        return {"mIOU": "{}".format(miou)}

    def per_class_metrics(self):
        """
        Computes IoU, Precision, Recall, F1-score for all classes.

        """
        if not hasattr(self._data, "load_empty"):
            raise Exception("Dataset is required for compute metrics")
        class_dict = (
            self._data._num_class_map_dict
            if self._data._num_class_map_dict
            else self._data._class_map_dict
        )
        mats, _ = model_eval(
            self._data, self.learn.model, class_dict, self._data._convertmap
        )

        mat_types = ["IoU", "Precision", "Recall", "F1-score"]

        mat = []
        for i in class_dict.values():
            for j in mat_types:
                if str(i) in mats[0].keys():
                    mat.append(mats[0][str(i)][j])

        matrix_1 = np.reshape(np.array(mat), (len(mats[0].keys()), 4))

        display(
            pd.DataFrame(
                matrix_1,
                index=[
                    self._data._class_map_dict.get(i)
                    for i, j in class_dict.items()
                    if str(j) in mats[0].keys()
                ],
                columns=mat_types,
            )
        )



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# error in training
# sparse_rcnn_r101_fpn_300_proposals_crop_mstrain_480-800_3x_coco.py, box AP=46.2

num_proposals = 300
num_stages = 6
model = dict(
    type="SparseRCNN",
    pretrained="torchvision://resnet101",
    backbone=dict(
        type="ResNet",
        depth=101,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=True,
        style="pytorch",
    ),
    neck=dict(
        type="FPN",
        in_channels=[256, 512, 1024, 2048],
        out_channels=256,
        start_level=0,
        add_extra_convs="on_input",
        num_outs=4,
    ),
    rpn_head=dict(
        type="EmbeddingRPNHead",
        num_proposals=num_proposals,
        proposal_feature_channel=256,
    ),
    roi_head=dict(
        type="SparseRoIHead",
        num_stages=num_stages,
        stage_loss_weights=[1] * num_stages,
        proposal_feature_channel=256,
        bbox_roi_extractor=dict(
            type="SingleRoIExtractor",
            roi_layer=dict(type="RoIAlign", output_size=7, sampling_ratio=2),
            out_channels=256,
            featmap_strides=[4, 8, 16, 32],
        ),
        bbox_head=[
            dict(
                type="DIIHead",
                num_classes=80,
                num_ffn_fcs=2,
                num_heads=8,
                num_cls_fcs=1,
                num_reg_fcs=3,
                feedforward_channels=2048,
                in_channels=256,
                dropout=0.0,
                ffn_act_cfg=dict(type="ReLU", inplace=True),
                dynamic_conv_cfg=dict(
                    type="DynamicConv",
                    in_channels=256,
                    feat_channels=64,
                    out_channels=256,
                    input_feat_shape=7,
                    act_cfg=dict(type="ReLU", inplace=True),
                    norm_cfg=dict(type="LN"),
                ),
                loss_bbox=dict(type="L1Loss", loss_weight=5.0),
                loss_iou=dict(type="GIoULoss", loss_weight=2.0),
                loss_cls=dict(
                    type="FocalLoss",
                    use_sigmoid=True,
                    gamma=2.0,
                    alpha=0.25,
                    loss_weight=2.0,
                ),
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    clip_border=False,
                    target_means=[0.0, 0.0, 0.0, 0.0],
                    target_stds=[0.5, 0.5, 1.0, 1.0],
                ),
            )
            for _ in range(num_stages)
        ],
    ),
    # training and testing settings
    train_cfg=dict(
        rpn=None,
        rcnn=[
            dict(
                assigner=dict(
                    type="HungarianAssigner",
                    cls_cost=dict(type="FocalLossCost", weight=2.0),
                    reg_cost=dict(type="BBoxL1Cost", weight=5.0),
                    iou_cost=dict(type="IoUCost", iou_mode="giou", weight=2.0),
                ),
                sampler=dict(type="PseudoSampler"),
                pos_weight=1,
            )
            for _ in range(num_stages)
        ],
    ),
    test_cfg=dict(_delete_=True, rpn=None, rcnn=dict(max_per_img=num_proposals)),
)

checkpoint = "https://download.openmmlab.com/mmdetection/v2.0/sparse_rcnn/sparse_rcnn_r101_fpn_300_proposals_crop_mstrain_480-800_3x_coco/sparse_rcnn_r101_fpn_300_proposals_crop_mstrain_480-800_3x_coco_20201223_023452-c23c3564.pth"



__version__ = "2.2.0.1"

from arcgis.auth.tools import LazyLoader

os = LazyLoader("os")
from arcgis import env
from arcgis import features

geoanalytics = LazyLoader("arcgis.geoanalytics")
geocoding = LazyLoader("arcgis.geocoding")
geometry = LazyLoader("arcgis.geometry")
geoprocessing = LazyLoader("arcgis.geoprocessing")
network = LazyLoader("arcgis.network")
raster = LazyLoader("arcgis.raster")
realtime = LazyLoader("arcgis.realtime")
schematics = LazyLoader("arcgis.schematics")
mapping = LazyLoader("arcgis.mapping")
apps = LazyLoader("arcgis.apps")

if not os.environ.get("DISABLE_ARCGIS_LEARN", None) == "1":
    learn = LazyLoader("arcgis.learn")

from arcgis.gis import GIS
from arcgis.features.analysis import (
    calculate_density,
    find_hot_spots,
    find_outliers,
    find_point_clusters,
    interpolate_points,
    summarize_center_and_dispersion,
    connect_origins_to_destinations,
    create_buffers,
    create_drive_time_areas,
    find_nearest,
    plan_routes,
    enrich_layer,
    choose_best_facilities,
    create_viewshed,
    create_watersheds,
    derive_new_locations,
    find_centroids,
    find_existing_locations,
    find_similar_locations,
    trace_downstream,
    create_route_layers,
    dissolve_boundaries,
    extract_data,
    generate_tessellation,
    merge_layers,
    overlay_layers,
    aggregate_points,
    join_features,
    summarize_center_and_dispersion,
    summarize_nearby,
    summarize_within,
)
from .geocoding import geocode


try:
    import pandas as pd
    from arcgis.features.geo import GeoAccessor
    from arcgis.features.geo import GeoSeriesAccessor


except ImportError as e:
    pass

try:
    # register with dask
    from .features.geo import _dask  # noqa
except Exception as e:
    pass


__all__ = [
    "GIS",
    "geocode",
    "features",
    "geoanalytics",
    "geocoding",
    "geometry",
    "geoprocessing",
    "network",
    "notebook",
    "raster",
    "apps",
    "realtime",
    "schematics",
    "mapping",
    "learn",
    "aggregate_points",
    "calculate_density",
    "connect_origins_to_destinations",
    "create_buffers",
    "create_drive_time_areas",
    "create_route_layers",
    "create_viewshed",
    "create_watersheds",
    "derive_new_locations",
    "dissolve_boundaries",
    "enrich_layer",
    "extract_data",
    "find_existing_locations",
    "find_hot_spots",
    "find_nearest",
    "find_similar_locations",
    "find_centroids",
    "interpolate_points",
    "join_features",
    "merge_layers",
    "overlay_layers",
    "plan_routes",
    "summarize_nearby",
    "summarize_center_and_dispersion",
    "summarize_within",
    "trace_downstream",
]


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "widgets/js/dist",
            "dest": "arcgis",
            "require": "arcgis/extension",
        }
    ]



"""

Contains an internal GP method used to calculate Credit Usage for Tool
This is a private method and could change without warning. Do not use.

"""
import json
import arcgis as _arcgis
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool


def _estimate_credits(task, parameters, gis=None):
    """
    Estimates the number of credits a spatial analysis operation will take.

    =======================     ====================================================================
    **Parameter**                **Description**
    -----------------------     --------------------------------------------------------------------
    task                        Required String. The name of the analysis tool.
    -----------------------     --------------------------------------------------------------------
    parameters                  Required String.  The input parameters for the tool.
    -----------------------     --------------------------------------------------------------------
    gis                         Optional GIS.  The enterprise connection object.
    =======================     ====================================================================

    :return: Float value indicating the maximum cost

    """

    if gis is None and _arcgis.env.active_gis:
        gis = _arcgis.env.active_gis
    elif gis is None and _arcgis.env.active_gis is None:
        raise Exception("A GIS must be provided and/or set as active.")
    if gis.version >= [6, 4] and gis._portal.is_arcgisonline:
        url = gis.properties["helperServices"]["creditEstimation"]["url"]
        gptask = "EstimateCredits"
        url = "{base}/{gptask}/execute".format(base=url, gptask=gptask)
        params = {
            "f": "json",
            "taskName": task,
            "taskParameters": json.dumps(parameters),
        }
        kwargs = locals()
        param_db = {
            "task": (str, "taskName"),
            "parameters": (str, "taskParameters"),
            "credit_estimate": (str, "creditEstimate"),
        }
        return_values = [
            {"name": "credit_estimate", "display_name": "creditEstimate", "type": str},
        ]
        res = _execute_gp_tool(
            gis,
            gptask,
            kwargs,
            param_db,
            return_values,
            False,
            url,
            webtool=True,
            add_token=False,
        )
        if "cost" in res:
            return float(res["cost"])
        elif "maximumCost" in res:
            return float(res["maximumCost"])
        return res
    return



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# faster_rcnn_regnetx-3.2GF_fpn_mstrain_3x_coco.py, box AP=42.2

_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"
model = dict(
    pretrained="open-mmlab://regnetx_3.2gf",
    backbone=dict(
        _delete_=True,
        type="RegNet",
        arch="regnetx_3.2gf",
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        norm_eval=True,
        style="pytorch",
    ),
    neck=dict(
        type="FPN", in_channels=[96, 192, 432, 1008], out_channels=256, num_outs=5
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/regnet/faster_rcnn_regnetx-3.2GF_fpn_mstrain_3x_coco/faster_rcnn_regnetx-3.2GF_fpn_mstrain_3x_coco_20200520_224253-bf85ae3e.pth"



"""
Affine transformation functions
"""
from __future__ import division
import math
from . import Point, Polygon, Polyline
from . import MultiPoint, Geometry

GEOM_TYPES = (Point, Polygon, Polyline, MultiPoint)
__all__ = ["scale", "rotate", "skew", "translate"]


# -------------------------------------------------------------------------
def scale(geom: Geometry, *scale_factor: int):
    """
    Create a scaling transform from a scalar value (float)

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    geom                Required class:`~arcgis.geometry.Geometry object or dictionary.
    ---------------     --------------------------------------------------------------------
    scale_factor        Required int. A scalar value will scale in both dimensions equally
    ===============     ====================================================================

    :return:
        Dictionary or class:`~arcgis.geometry.Geometry` object
    """
    import numpy as np

    if len(scale_factor) == 1:
        sy = sx = scale_factor[0]
    elif len(scale_factor) > 1:
        sx = scale_factor[0]
        sy = scale_factor[1]
    else:
        return
    A = np.matrix([[sx, 0.0], [0.0, sy]])
    if isinstance(geom, dict) or isinstance(geom, GEOM_TYPES):
        if "x" in geom and "y" in geom:  # translates point
            matrix = np.matrix([[geom["x"]], [geom["y"]]])
            val = (A * matrix).tolist()
            geom["x"] = val[0][0]
            geom["y"] = val[1][0]
            return geom
        elif "rings" in geom or "paths" in geom:  # translates lines/polygons
            rings = []
            if "paths" in geom:
                coords = geom["paths"]
            else:
                coords = geom["rings"]
            for part in coords:
                x_coords = [pt[0] for pt in part]
                y_coords = [pt[1] for pt in part]
                matrix = np.matrix([x_coords, y_coords])
                vals = (A * matrix).tolist()
                rings.append(list(zip(vals[0], vals[1])))
            if "paths" in geom:
                geom["paths"] = rings
            else:
                geom["rings"] = rings
            return geom
        elif "points" in geom:  # translates Multipoint
            for pt in geom["points"]:
                matrix = np.matrix([[pt[0]], [pt[1]]])
                val = (A * matrix).tolist()
                pt[0] = val[0][0]
                pt[1] = val[1][0]
            return geom
        else:
            return None
    return None


# -------------------------------------------------------------------------
def rotate(geom: Geometry, theta: float):
    """
    Rotates a geometry counter-clockwise by some degree theta

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    geom                Required class:`~arcgis.geometry.Geometry object or dictionary.
    ---------------     --------------------------------------------------------------------
    theta               Required angle of rotation
    ===============     ====================================================================

    :return:
        Dictionary or class:`~arcgis.geometry.Geometry` object
    """
    import numpy as np

    A = np.matrix(
        [[math.cos(theta), -1 * math.sin(theta)], [math.sin(theta), math.cos(theta)]]
    )
    if isinstance(geom, dict) or isinstance(geom, GEOM_TYPES):
        if "x" in geom and "y" in geom:  # translates point
            matrix = np.matrix([[geom["x"]], [geom["y"]]])
            val = (A * matrix).tolist()
            geom["x"] = val[0][0]
            geom["y"] = val[1][0]
            return geom
        elif "rings" in geom or "paths" in geom:  # translates lines/polygons
            rings = []
            if "paths" in geom:
                coords = geom["paths"]
            else:
                coords = geom["rings"]
            for part in coords:
                x_coords = [pt[0] for pt in part]
                y_coords = [pt[1] for pt in part]
                matrix = np.matrix([x_coords, y_coords])
                vals = (A * matrix).tolist()
                rings.append(list(zip(vals[0], vals[1])))
            if "paths" in geom:
                geom["paths"] = rings
            else:
                geom["rings"] = rings
            return geom
        elif "points" in geom:  # translates Multipoint
            for pt in geom["points"]:
                matrix = np.matrix([[pt[0]], [pt[1]]])
                val = (A * matrix).tolist()
                pt[0] = val[0][0]
                pt[1] = val[1][0]
            return geom
        else:
            return None
    return None


# -------------------------------------------------------------------------
def skew(geom: Geometry, x_angle: float = 0, y_angle: float = 0):
    """
    Create a skew transform along one or both axes.

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    geom                Required class:`~arcgis.geometry.Geometry object or dictionary.
    ---------------     --------------------------------------------------------------------
    x_angle             Required angle to skew in the x-coordinates
    ---------------     --------------------------------------------------------------------
    y_angle             Required angle to skew in the y-coordinates
    ===============     ====================================================================

    :return:
        Dictionary or class:`~arcgis.geometry.Geometry` object
    """
    import numpy as np

    A = np.matrix([[1, math.tan(x_angle)], [math.tan(y_angle), 1]])
    if isinstance(geom, dict) or isinstance(geom, GEOM_TYPES):
        if "x" in geom and "y" in geom:  # translates point
            matrix = np.matrix(
                [
                    [geom["x"]],
                    [geom["y"]],
                ]
            )
            val = (A * matrix).tolist()
            geom["x"] = val[0][0]
            geom["y"] = val[1][0]
            return geom
        elif "rings" in geom or "paths" in geom:  # translates lines/polygons
            rings = []
            if "paths" in geom:
                coords = geom["paths"]
            else:
                coords = geom["rings"]
            for part in coords:
                x_coords = [pt[0] for pt in part]
                y_coords = [pt[1] for pt in part]
                matrix = np.matrix([x_coords, y_coords])
                vals = (A * matrix).tolist()
                rings.append(list(zip(vals[0], vals[1])))
            if "paths" in geom:
                geom["paths"] = rings
            else:
                geom["rings"] = rings
            return geom
        elif "points" in geom:  # translates Multipoint
            for pt in geom["points"]:
                matrix = np.matrix(
                    [
                        [pt[0]],
                        [pt[1]],
                    ]
                )
                val = (A * matrix).tolist()
                pt[0] = val[0][0]
                pt[1] = val[1][0]
            return geom
        else:
            return None


# -------------------------------------------------------------------------
def translate(geom: Geometry, x_offset: float, y_offset: float):
    """
    Moves a geometry by some distance

    ===============     ====================================================================
    **Parameter**        **Description**
    ---------------     --------------------------------------------------------------------
    geom                Required class:`~arcgis.geometry.Geometry object or dictionary.
    ---------------     --------------------------------------------------------------------
    x_offset            Required int. Distance to move coordinates in the x-direction
    ---------------     --------------------------------------------------------------------
    y_offset            Required int. Distance to move coordinates in the y-direction
    ===============     ====================================================================

    :return:
        Dictionary or class:`~arcgis.geometry.Geometry` object
    """
    import numpy as np

    A = np.matrix([[1, 0, x_offset], [0, 1, y_offset], [0, 0, 1]])
    if isinstance(geom, dict) or isinstance(geom, GEOM_TYPES):
        if "x" in geom and "y" in geom:  # translates point
            matrix = np.matrix([[geom["x"]], [geom["y"]], [1]])
            val = (A * matrix).tolist()
            geom["x"] = val[0][0]
            geom["y"] = val[1][0]
            return geom
        elif "rings" in geom or "paths" in geom:  # translates lines/polygons
            rings = []
            if "paths" in geom:
                coords = geom["paths"]
            else:
                coords = geom["rings"]
            for part in coords:
                x_coords = [pt[0] for pt in part]
                y_coords = [pt[1] for pt in part]
                matrix = np.matrix([x_coords, y_coords, [1] * len(y_coords)])
                vals = (A * matrix).tolist()
                rings.append(list(zip(vals[0], vals[1])))
            if "paths" in geom:
                geom["paths"] = rings
            else:
                geom["rings"] = rings
            return geom
        elif "points" in geom:  # translates Multipoint
            for pt in geom["points"]:
                matrix = np.matrix([[pt[0]], [pt[1]], [1]])
                val = (A * matrix).tolist()
                pt[0] = val[0][0]
                pt[1] = val[1][0]
            return geom
        else:
            return None


if __name__ == "__main__":
    shapes = [
        {
            "rings": [
                [
                    [-97.06138, 32.837],
                    [-97.06133, 32.836],
                    [-97.06124, 32.834],
                    [-97.06127, 32.832],
                    [-97.06138, 32.837],
                ],
                [
                    [-97.06326, 32.759],
                    [-97.06298, 32.755],
                    [-97.06153, 32.749],
                    [-97.06326, 32.759],
                ],
            ],
            "spatialReference": {"wkid": 4326},
        },
        {"x": 50, "y": 60},
        {
            "paths": [
                [
                    [-97.06138, 32.837],
                    [-97.06133, 32.836],
                    [-97.06124, 32.834],
                    [-97.06127, 32.832],
                ],
                [[-97.06326, 32.759], [-97.06298, 32.755]],
            ],
            "spatialReference": {"wkid": 4326},
        },
        {
            "points": [
                [-97.06138, 32.837],
                [-97.06133, 32.836],
                [-97.06124, 32.834],
                [-97.06127, 32.832],
            ],
            "spatialReference": {"wkid": 4326},
        },
    ]
    # translate Test
    print("translate test")
    for shape in shapes:
        # val_dict = translate(shape,-10, 10)
        val_geom = translate(Geometry(shape), -10, 10)
        print(val_geom)
        print()
    print("rotate shapes")
    for shape in shapes:
        val_geom = rotate(geom=Geometry(shape), theta=20)
        print(val_geom, val_geom.type)
        del shape
    print("scale factor shapes")
    for shape in shapes:
        val_geom = scale(Geometry(shape), 10)
        val_geom = scale(Geometry(shape), 10, 20)
    print("skew test")
    for shape in shapes:  # Skew test
        val_geom = skew(geom=Geometry(shape), x_angle=45, y_angle=-20)
    print()



"""
StoryMap Implementation
"""
from .storymap import JournalStoryMap
from .story import StoryMap, Themes
from .story_content import (
    Image,
    Video,
    Audio,
    Embed,
    Text,
    Button,
    Map,
    Sidecar,
    Gallery,
    Timeline,
    Swipe,
    TextStyles,
    Scales,
    MapTour,
)

__all__ = [
    "JournalStoryMap",
    "StoryMap",
]



import re
import os
import sys
import copy
import types
import shutil
import random
import logging
import warnings
import traceback
import pandas as pd
from functools import partial
from .._utils.common import always_warn

HAS_FASTAI = True
try:
    import torch
    import arcgis
    from fastai.text import (
        TextList,
        TextClasDataBunch,
        pad_collate,
        TextDataBunch,
        SortishSampler,
        SortSampler,
        ItemList,
        ItemBase,
        Text,
    )
    from fastai.data_block import CategoryList, MultiCategoryList
    from ._seq2seq_utils import SequenceToSequenceTextList, teacher_forcing_tfm
    from .text_transforms import (
        TransformerNERDataset,
        TransformerNERDataBunch,
        process_text,
    )
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

HAS_BEAUTIFULSOUP = True
try:
    from bs4 import BeautifulSoup
except:
    HAS_BEAUTIFULSOUP = False
else:
    warnings.filterwarnings("ignore", category=UserWarning, module="bs4")

HAS_NUMPY = True
try:
    import numpy as np

    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
except:
    HAS_NUMPY = False


max_len = 100


def _raise_fastai_exception(exception):
    error_message = (
        f"{exception}\n\nThis module requires fastai, PyTorch and transformers as its dependencies.\n"
        "Install them using - 'conda install -c esri -c fastai -c pytorch arcgis=1.8.2 "
        "scikit-image=0.15.0 pillow=6.2.2 libtiff=4.0.10 fastai=1.0.60 pytorch=1.4.0 "
        "torchvision=0.5.0 scikit-learn=0.23.1 --no-pin'"
        "\n'conda install gdal=2.3.3'"
        "\n'pip install transformers==3.3.0'"
    )
    raise Exception(error_message)


# Overriding show_text_xys function of TextList to display the dataframe in desired fashion
def show_text_xys(self, xs, ys, max_len: int = max_len):
    "Show the `xs` (inputs) and `ys` (targets). `max_len` is the maximum number of tokens displayed."
    if not HAS_NUMPY:
        raise Exception("This module requires numpy.")

    from IPython.display import display

    names = ["idx", "text"] if self._is_lm else ["text", "target"]
    items = []
    for i, (x, y) in enumerate(zip(xs, ys)):
        txt_x = " ".join(x.text.split(" ")[:max_len]) if max_len is not None else x.text
        items.append([i, txt_x] if self._is_lm else [txt_x, y])
    items = np.array(items)
    df = pd.DataFrame({n: items[:, i] for i, n in enumerate(names)}, columns=names)
    dataframe_style = (
        df.style.set_table_styles([dict(selector="th", props=[("text-align", "left")])])
        .set_properties(**{"text-align": "left"})
        .hide(axis="index")
    )
    display(dataframe_style)


# Overriding show_text_xyzs function of TextList to display the dataframe in desired fashion
def show_text_xyzs(self, xs, ys, zs, max_len: int = max_len):
    "Show `xs` (inputs), `ys` (targets) and `zs` (predictions). `max_len` is the maximum number of tokens displayed."
    if not HAS_NUMPY:
        raise Exception("This module requires numpy.")

    from IPython.display import display

    items, names = [], ["text", "target", "prediction"]
    for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
        txt_x = " ".join(x.text.split(" ")[:max_len]) if max_len is not None else x.text
        items.append([txt_x, y, z])
    items = np.array(items)
    df = pd.DataFrame({n: items[:, i] for i, n in enumerate(names)}, columns=names)
    dataframe_style = (
        df.style.set_table_styles([dict(selector="th", props=[("text-align", "left")])])
        .set_properties(**{"text-align": "left"})
        .hide(axis="index")
    )
    display(dataframe_style)


def read_file(path):
    filename, file_extension = os.path.splitext(path)
    if file_extension == ".csv":
        return pd.read_csv(path, dtype="str")
    else:
        return pd.read_csv(path, sep="\t", dtype="str")


def save_data_in_model_metrics_html(text, path, model_characteristics_folder):
    file_path = os.path.join(path, "model_metrics.html")
    if os.path.exists(file_path):
        with open(file_path, mode="a+", encoding="utf-8") as fp:
            fp.write(text)

    folder_path = os.path.join(path, model_characteristics_folder)
    if os.path.exists(folder_path):
        with open(
            os.path.join(folder_path, "sample_results.html"), mode="w", encoding="utf-8"
        ) as fp:
            fp.write(text)


def copy_metrics(source, target, folder_name):
    source_file_path = os.path.join(source, "model_metrics.html")
    target_file_path = os.path.join(target, "model_metrics.html")
    if os.path.exists(source_file_path):
        if os.path.exists(target_file_path):
            os.remove(target_file_path)
        _ = shutil.copyfile(source_file_path, target_file_path)

    source_folder_path = os.path.join(source, folder_name)
    target_folder_path = os.path.join(target, folder_name)
    if os.path.exists(source_folder_path):
        if os.path.exists(target_folder_path):
            shutil.rmtree(target_folder_path)
        _ = shutil.copytree(source_folder_path, target_folder_path)


def extract_entities(tokens, labels):
    prev_label, token_list, entities = labels[0], [tokens[0]], []
    prev_label = prev_label.split("-")[-1]
    for token_index, (token, label) in enumerate(list(zip(tokens[1:], labels[1:]))):
        label = label.split("-")[-1]
        if label == "O":
            if prev_label != "O":
                entities.append((token_list, prev_label))
            token_list, prev_label = list(), label

        if prev_label == label:
            token_list.append(token)
        else:
            if prev_label != "O":
                entities.append((token_list, prev_label))
            token_list = list()
            prev_label = label
            token_list.append(token)
    if token_list:
        entities.append((token_list, prev_label))
    return entities


class TextDataObject:
    def __init__(self, task):
        self.emd = {}
        self.emd_path = None

        self._bs = None
        self._is_empty = True
        self._train_df = None
        self._valid_df = None
        self._text_cols = None
        self._label_cols = list()
        self._databunch = None
        self._training_indexes = list()
        self._task = task
        self._backbone = None

        self.databunch_kwargs = {"num_workers": 0} if sys.platform == "win32" else {}
        self.databunch_kwargs["pin_memory"] = True
        if (
            getattr(arcgis.env, "_processorType", "") == "GPU"
            and torch.cuda.is_available()
        ):
            self.databunch_kwargs["device"] = torch.device("cuda")
        elif getattr(arcgis.env, "_processorType", "") == "CPU":
            self.databunch_kwargs["device"] = torch.device("cpu")
        else:
            self.databunch_kwargs["device"] = (
                torch.device("cuda")
                if torch.cuda.is_available()
                else torch.device("cpu")
            )

    @classmethod
    def prepare_data_for_entity_recognition(
        cls,
        tokens_collection,
        tags_collection,
        address_tag,
        unique_tags,
        seed=42,
        batch_size=8,
        val_split_pct=0.1,
        label2id=None,
    ):
        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        text_data = cls("ner")
        random.seed(seed)
        # temp = list(zip(tokens_collection, tags_collection))
        # random.shuffle(temp)
        # tokens_collection, tags_collection = zip(*temp)

        random.Random(seed).shuffle(tokens_collection)
        random.Random(seed).shuffle(tags_collection)

        # used when creating databunch for ner
        split_index = int((1 - val_split_pct) * len(tokens_collection))

        text_data._bs = batch_size
        text_data._unique_tags = unique_tags
        text_data._address_tag = address_tag
        text_data._label2id = (
            label2id if label2id else {x: index for index, x in enumerate(unique_tags)}
        )

        text_data._train_tags = tags_collection[:split_index]
        text_data._train_tokens = tokens_collection[:split_index]

        text_data._valid_tags = tags_collection[split_index:]
        text_data._valid_tokens = tokens_collection[split_index:]
        if text_data._address_tag not in text_data._unique_tags:
            logging.warning(
                "No Address tag found in your data.\n\
            1. If your data has an address field, pass your address field name as address tag in class mapping \n\
            e.g. - data=prepare_data(dataset_type=ds_type,path=training_data_folder,\n\t\t\t\
                class_mapping={address_tag:address_field_name})\n\
            2. Else no action is required, if your data does not have any address information."
            )

        return text_data

    @classmethod
    def prepare_data_for_classification(
        cls,
        data,
        text_cols,
        label_cols,
        train_file="train.csv",
        valid_file=None,
        val_split_pct=0.1,
        seed=42,
        batch_size=8,
        process_labels=False,
        remove_html_tags=False,
        remove_urls=False,
        **kwargs,
    ):
        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        text_data = cls("classification")
        if not os.path.exists(data):
            raise Exception(f"Provided data directory - {data}, does not exists")

        training_file_path = os.path.join(data, train_file)

        if not os.path.exists(training_file_path):
            raise Exception(
                f"Provided data directory does not contain {train_file} file"
            )

        train_df = read_file(training_file_path)
        train_df = cls._preprocess_df(
            train_df,
            text_cols,
            label_cols,
            process_labels,
            remove_html_tags,
            remove_urls,
        )

        random.seed(seed)

        if valid_file is not None and os.path.exists(os.path.join(data, valid_file)):
            validation_file_exists = True
        else:
            validation_file_exists = False

        if validation_file_exists:
            valid_df = read_file(os.path.join(data, valid_file))
            valid_df = cls._preprocess_df(
                valid_df,
                text_cols,
                label_cols,
                process_labels,
                remove_html_tags,
                remove_urls,
            )
        else:
            if len(label_cols) == 1 and kwargs.get("stratify") != False:
                label_col = label_cols[0]
                from sklearn.model_selection import train_test_split

                x, y = train_df[text_cols], train_df[label_col]
                unique_labels = y.value_counts()[y.value_counts() == 1].index.tolist()

                for (
                    label
                ) in unique_labels:  # duplicating datapoints with unique classes.
                    idx = y[y == label].index.tolist()[0]
                    train_df = train_df.append(train_df.iloc[idx])
                train_df.reset_index(drop=True, inplace=True)
                x, y = train_df[text_cols], train_df[label_col]
                X_train, X_test, y_train, y_test = train_test_split(
                    x, y, test_size=val_split_pct, stratify=y
                )
                train_df = pd.concat([X_train, y_train], axis=1)
                valid_df = pd.concat([X_test, y_test], axis=1)
                temp_df = pd.DataFrame()
            else:
                validation_indexes = random.sample(
                    range(train_df.shape[0]), round(val_split_pct * train_df.shape[0])
                )
                training_indexes = list(
                    set([i for i in range(train_df.shape[0])]) - set(validation_indexes)
                )

                temp_df = copy.deepcopy(train_df)
                train_df = temp_df.loc[training_indexes]
                valid_df = temp_df.loc[validation_indexes]
            # Removing rows with empty strings in the text_cols from the training and validation data
            train_df[text_cols].replace("", np.nan, inplace=True)
            train_df.dropna(inplace=True)
            valid_df[text_cols].replace("", np.nan, inplace=True)
            valid_df.dropna(inplace=True)
            # Resetting dataframe indexes for training anf validation dataframe
            train_df.reset_index(drop=True, inplace=True)
            valid_df.reset_index(drop=True, inplace=True)
            del temp_df
        if len(label_cols) > 1:
            text_data.classes = label_cols
        else:
            text_data.classes = valid_df[label_cols[0]].unique().tolist()
        text_data._bs = batch_size
        text_data._text_cols = text_cols
        text_data._label_cols = label_cols
        text_data._train_df = train_df[[text_cols] + label_cols]
        text_data._valid_df = valid_df[[text_cols] + label_cols]
        text_data._training_indexes = range(0, train_df.shape[0])

        return text_data

    @classmethod
    def prepare_data_for_seq2seq(
        cls,
        data,
        text_cols,
        label_cols,
        train_file="train.csv",
        val_split_pct=0.1,
        seed=42,
        batch_size=8,
        process_labels=True,
        remove_html_tags=False,
        remove_urls=False,
    ):
        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        text_data = cls("sequence_translation")
        if not os.path.exists(data):
            raise Exception(f"Provided data directory - {data}, does not exists")

        training_file_path = os.path.join(data, train_file)

        if not os.path.exists(training_file_path):
            raise Exception(
                f"Provided data directory does not contain {train_file} file"
            )

        train_df = read_file(training_file_path)
        train_df = cls._preprocess_df(
            train_df,
            text_cols,
            label_cols,
            process_labels,
            remove_html_tags,
            remove_urls,
        )

        random.seed(seed)

        validation_indexes = random.sample(
            range(train_df.shape[0]), round(val_split_pct * train_df.shape[0])
        )
        training_indexes = list(
            set([i for i in range(train_df.shape[0])]) - set(validation_indexes)
        )

        temp_df = copy.deepcopy(train_df)
        train_df = temp_df.loc[training_indexes]
        valid_df = temp_df.loc[validation_indexes]
        # Removing rows with empty strings in the text_cols from the training and validation data
        train_df[text_cols].replace("", np.nan, inplace=True)
        train_df.dropna(inplace=True)
        valid_df[text_cols].replace("", np.nan, inplace=True)
        valid_df.dropna(inplace=True)
        # Resetting dataframe indexes for training anf validation dataframe
        train_df.reset_index(drop=True, inplace=True)
        valid_df.reset_index(drop=True, inplace=True)
        del temp_df
        text_data._bs = batch_size
        text_data._text_cols = text_cols
        text_data._label_cols = label_cols
        text_data._train_df = train_df[[text_cols] + label_cols]
        text_data._valid_df = valid_df[[text_cols] + label_cols]
        text_data.val_split_pct = val_split_pct
        text_data._training_indexes = range(0, train_df.shape[0])
        text_data._model_type = None
        return text_data

    def get_databunch(self):
        if self._is_empty:
            return None
        else:
            return self._databunch

    def _prepare_databunch(
        self, tokenizer, vocab=None, pad_first=True, pad_idx=0, **kwargs
    ):
        """
        Wrapper to create fastai TextDataBunch Object
        """
        if not HAS_FASTAI:
            _raise_fastai_exception(import_exception)

        logger = kwargs.get("logger")
        classes = kwargs.get("classes", None)
        backbone = kwargs.get("backbone")
        if classes:
            label_cols = self._label_cols
            unique_labels = (
                list(self._train_df[label_cols[0]].unique())
                if len(label_cols) == 1
                else label_cols
            )
            if set(classes) != set(unique_labels):
                error_message = (
                    f"Looks like the - `{backbone}` is fine-tuned on the following labels - "
                    f"{classes} and your data consists of the following labels - `{unique_labels}`."
                    f"\nPlease use a base model of - `{backbone}` and fine-tune it on your data or use a "
                    f"model which is fine-tuned on a data having same labels as - `{unique_labels}`"
                )
                raise Exception(error_message)

        if logger:
            logger.info(f"Preparing databunch for task type - {self._task}")
        if self._task == "classification":
            self._databunch = TextClasDataBunch.from_df(
                ".",
                train_df=self._train_df,
                valid_df=self._valid_df,
                tokenizer=tokenizer,
                vocab=vocab,
                include_bos=False,
                include_eos=False,
                text_cols=self._text_cols,
                label_cols=self._label_cols,
                bs=self._bs,
                collate_fn=partial(pad_collate, pad_first=pad_first, pad_idx=pad_idx),
                classes=classes,
                **self.databunch_kwargs,
            )

            TextList.show_xys = types.MethodType(show_text_xys, TextList)
            TextList.show_xyzs = types.MethodType(show_text_xyzs, TextList)
        elif self._task == "ner":
            model_type, seq_length = kwargs["model_type"], kwargs["seq_len"]
            dl_kwargs = {"pin_memory": self.databunch_kwargs.get("pin_memory")}
            device, num_workers = (
                self.databunch_kwargs["device"],
                self.databunch_kwargs.get("num_workers"),
            )
            if num_workers:
                dl_kwargs["num_workers"] = num_workers

            train_ds = TransformerNERDataset(
                self._train_tokens,
                self._train_tags,
                tokenizer,
                self._label2id,
                model_type=model_type,
                seq_length=seq_length,
            )
            valid_ds = TransformerNERDataset(
                self._valid_tokens,
                self._valid_tags,
                tokenizer,
                self._label2id,
                model_type=model_type,
                seq_length=seq_length,
            )

            self._databunch = TransformerNERDataBunch.create(
                train_ds,
                valid_ds,
                tokenizer,
                bs=self._bs,
                device=device,
                seq_len=seq_length,
                **dl_kwargs,
            )
        else:
            raise Exception(
                f"Wrong task - {self._task} selected. Allowed values are 'ner', 'classification'"
            )

        self._is_empty = False
        self._backbone = kwargs.get("backbone")

    def _prepare_seq2seq_databunch(
        self, transformer_processor, pad_first, pad_idx, **kwargs
    ):
        """
        Wrapper to create fastai TextDataBunch Object
        """
        self._model_type = kwargs.get("model_type")

        if not HAS_FASTAI:
            return
        dl_tfms = None
        if self._model_type in ["t5", "bart", "marian"]:
            dl_tfms = teacher_forcing_tfm
        data = (
            SequenceToSequenceTextList.from_df(
                self._train_df, cols=self._text_cols, processor=transformer_processor
            )
            .split_by_rand_pct(valid_pct=self.val_split_pct)
            .label_from_df(
                cols=self._label_cols,
                label_cls=TextList,
                processor=transformer_processor,
            )
            .databunch(
                bs=self._bs,
                pad_first=pad_first,
                pad_idx=pad_idx,
                dl_tfms=dl_tfms,
                **self.databunch_kwargs,
            )
        )

        self._is_empty = False
        self._databunch = data

        self._backbone = kwargs.get("backbone")

    @staticmethod
    def _preprocess_df(
        dataframe,
        text_cols,
        label_cols,
        process_labels,
        remove_html_tags=False,
        remove_urls=False,
    ):
        """
        Do some pre-processing on the dataframe columns
        """
        dataframe[text_cols].fillna("", inplace=True)
        dataframe[text_cols] = dataframe.apply(
            lambda row: TextDataObject.process_text(
                row[text_cols], remove_html_tags, remove_urls
            ),
            axis=1,
        )
        if process_labels:
            for label in label_cols:
                dataframe[label] = dataframe.apply(
                    lambda row: TextDataObject.process_text(
                        row[label], remove_html_tags, remove_urls
                    ),
                    axis=1,
                )

        return dataframe

    @staticmethod
    def process_text(
        text,
        remove_html_tags=False,
        remove_urls=False,
        remove_space_before_after_spl_char=False,
    ):
        """
        Perform some basic cleanup like removing HTML tags, removing urls,
        removing spaces before and after special characters
        converting multiple white spaces to single white space
        """
        if remove_urls:
            text = re.sub(
                r"\b(?:(?:https?|ftp)://)?\w[\w-]*(?:\.[\w-]+)+\S*", " ", text
            )

        # text = re.sub(r'<.*?>', '', text)
        if remove_html_tags:
            if not HAS_BEAUTIFULSOUP:
                raise Exception("This module requires BeautifulSoup.")
            text = BeautifulSoup(text, "html.parser").get_text(
                separator=" ", strip=True
            )

        if remove_space_before_after_spl_char:
            text = re.sub(" ([@.#$/:-]) ?", r"\1", text)

        if any([remove_html_tags, remove_urls, remove_space_before_after_spl_char]):
            text = re.sub(" +", " ", text)
        return text.strip()

    def show_batch(self, rows=5, max_len=max_len):
        """
        Shows a batch of dataframe prepared without applying transforms.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional integer. Number of rows in the
                                dataframe to be shown on the function call.
                                The default value is `5`.
        ---------------------   -------------------------------------------
        max_len                 Optional integer. Maximum number of tokens to be
                                shown for the text field (source column) of the
                                dataframe. The default value is `100`.
        =====================   ===========================================
        """
        if self._task == "classification":
            return self._classification_show_batch(rows=rows, max_len=max_len)
        elif self._task == "ner":
            return self._ner_show_batch(rows=rows)
        elif self._task == "sequence_translation":
            return self._classification_show_batch(rows=rows, max_len=max_len)
        else:
            raise Exception(
                f"Wrong task - {self._task} selected. Allowed values are 'ner', 'classification'"
            )

    def _ner_show_batch(self, rows=5):
        results = []
        indexes = list(range(len(self._train_tokens)))
        random.shuffle(indexes)
        rows = min(len(indexes), rows)
        batch_tokens = [self._train_tokens[i] for i in indexes[:rows]]
        batch_labels = [self._train_tags[i] for i in indexes[:rows]]
        for index, (tokens, labels) in enumerate(zip(batch_tokens, batch_labels)):
            entity_dict = dict()
            entities = extract_entities(tokens, labels)
            text = process_text(" ".join(tokens))
            entity_dict["Text"] = text
            _ = [
                entity_dict.setdefault(x[1], []).append(process_text(" ".join(x[0])))
                for x in entities
                if x[0]
            ]
            results.append(entity_dict)
        df = pd.DataFrame(
            results,
        )
        df.fillna("", inplace=True)
        return df

    def _classification_show_batch(self, rows=5, max_len=max_len):
        processed_data = []
        rows = min(len(self._training_indexes), rows)
        random_batch = random.sample(self._training_indexes, rows)
        # rows = min(len(self._train_df), rows)
        # random_batch = np.random.randint(0,self._train_df.index.max(),rows)
        dataframe = self._train_df.loc[random_batch]
        for idx, item in dataframe.iterrows():
            if len(self._label_cols) > 1:
                target = ";".join(
                    [
                        column
                        for column in self._label_cols
                        if int(getattr(item, column))
                    ]
                )
            else:
                target = getattr(item, self._label_cols[0])
            source = getattr(item, self._text_cols)
            if max_len is not None:
                source = " ".join(source.split(" "))[:max_len]
            processed_data.append([source, target])

            dataframe = pd.DataFrame(processed_data, columns=["source", "target"])

        return (
            dataframe.style.set_table_styles(
                [dict(selector="th", props=[("text-align", "left")])]
            )
            .set_properties(**{"text-align": "left"})
            .hide(axis="index")
        )

    def create_empty_object_for_ner(
        self, entities, address_tag, label2id, batch_size=4
    ):
        self._bs = batch_size
        self._unique_tags = entities
        self._address_tag = address_tag
        self._label2id = label2id
        self._train_tags, self._train_tokens = [], []
        self._valid_tags, self._valid_tokens = [], []

    def create_empty_object_for_classification(
        self, text_cols, label_cols, classes, is_multilabel=False
    ):
        self._text_cols = text_cols
        self._label_cols = label_cols

        text_list = TextList([], ignore_empty=True).split_by_idx([])
        if is_multilabel:
            self._databunch = text_list.label_const(
                0, label_cls=MultiCategoryList, classes=classes
            ).databunch()
        else:
            self._databunch = text_list.label_const(
                0, label_cls=CategoryList, classes=classes
            ).databunch()

        self._is_empty = False

    def create_empty_seq2seq_data(self, text_cols, label_cols):
        self._text_cols = text_cols
        self._label_cols = label_cols

        self._databunch = (
            SequenceToSequenceTextList([]).split_none().label_const().databunch()
        )
        self._is_empty = False

        # self._train_df = self._create_empty_df(batch_size, [text_cols] + label_cols)
        # self._valid_df = self._create_empty_df(batch_size, [text_cols] + label_cols)

    # @staticmethod
    # def _create_empty_df(batch_size, columns):
    #     l1 = ["" for x in columns]
    #     l2 = [l1 for x in range(batch_size)]
    #     return pd.DataFrame(l2, columns=columns)



import torch
import torch.nn as nn
import torchvision.models as models
from torch.nn.utils.rnn import pack_padded_sequence
from fastai.vision import create_body
from fastai.basic_train import Learner
import torch.nn.functional as F
from fastai.basic_train import LearnerCallback
import random
from functools import partial
from .._utils.image_captioning_data import show_image_and_text
from fastai.vision import open_image
import matplotlib.pyplot as plt
from fastai.callbacks.hooks import model_sizes
import numpy as np
from fastai.callback import Callback
from fastai.torch_core import add_metrics
from fastai.text import Counter, exp
from .._utils.image_captioning_data import BeamSearchAttention as BeamSearch
from random import choice
from .._utils.image_captioning_data import normalize
from fastai.torch_core import split_model_idx, flatten_model
from pathlib import Path
import shutil
import os
import sys
import tempfile
import fasttext
import fasttext.util


EPS = 1e-5


class EncoderAttention(nn.Module):
    def __init__(self, backbone, cut=None, pretrained=True):
        """Load the pretrained backbone and replace top fc layer."""
        super().__init__()
        self.backbone = create_body(backbone, cut=cut, pretrained=pretrained)
        # Get number of channels of backbone.
        self.feature_size = model_sizes(self.backbone, size=(200, 200))[-1][1]

    def forward(self, images):
        """Extract feature vectors from input images."""
        features = self.backbone(images)
        B, C, H, W = features.shape
        features = features.permute(0, 2, 3, 1).view(B, H * W, C)
        return features.contiguous()

    def freeze(self):
        for p in self.parameters():
            p.requires_grad = False


class BahdanauAttention(nn.Module):
    def __init__(self, encoder_size, hidden_size, attention_size):
        super().__init__()
        self.W1 = nn.Linear(encoder_size, attention_size)
        self.W2 = nn.Linear(hidden_size, attention_size)
        self.V = nn.Linear(attention_size, 1)

    def forward(self, features, hidden):
        # features shape: B, spatial_size x spatial size, num_channels
        # hidden shape: B, hidden_size
        hidden = hidden.unsqueeze(1)
        feature_attention = self.W1(features)
        hidden_attention = self.W2(hidden)
        B, HW, C = features.shape
        C_att = feature_attention.shape[-1]
        B_hidden = hidden_attention.shape[0]

        # This will be required while doing beam search
        if not self.training and B_hidden != B:
            beam_size = int(B_hidden / B)
            feature_attention = feature_attention.unsqueeze(1).repeat(
                1, beam_size, 1, 1
            )
            feature_attention = feature_attention.view(B * beam_size, HW, C_att)
            features = features.unsqueeze(1).repeat(1, beam_size, 1, 1)
            features = features.view(B * beam_size, HW, C)

        score = torch.tanh(feature_attention + hidden_attention)
        attention_weights = F.softmax(self.V(score), dim=1)
        context_vector = attention_weights * features
        context_vector = context_vector.sum(dim=1)
        return context_vector, attention_weights


def select_nucleus(outp, p=0.5):
    # print("select_nuc", outp.shape)
    # import pdb; pdb.set_trace();
    outp = outp.squeeze(0)
    probs = F.softmax(outp, dim=-1)
    idxs = torch.argsort(probs, descending=True)
    res, cumsum = [], 0.0
    for idx in idxs:
        res.append(idx)
        cumsum += probs[idx]
        if cumsum > p:
            return idxs.new_tensor([choice(res)])


class DecoderAttention(nn.Module):
    def __init__(
        self,
        embed_size,
        feature_size,
        hidden_size,
        vocab_size,
        attention_size,
        max_seq_length=20,
        pretrained_embeddings=False,
        vocab=None,
        teacher_forcing=True,
        dropout=0.5,
    ):
        """
        Decoder will take in features and will generate
        captions from those. It defines the basic LSTM,
        attention and few linear layers.
        """
        super(DecoderAttention, self).__init__()
        self.vocab = vocab
        if pretrained_embeddings is not False:
            self.embed = self.load_embedding(pretrained_embeddings, vocab)
        else:
            self.embed = nn.Embedding(vocab_size, embed_size)

        self.lstm = nn.LSTMCell(2 * embed_size, hidden_size)
        self.linear = nn.Linear(hidden_size, vocab_size)
        self.embed_features = nn.Linear(feature_size, embed_size)
        self.hidden_init = nn.Linear(embed_size, hidden_size)
        self.c_init = nn.Linear(embed_size, hidden_size)
        self.attention = BahdanauAttention(embed_size, hidden_size, attention_size)
        self.out = nn.Linear(hidden_size, vocab_size)
        self.max_seg_length = max_seq_length
        self.teacher_forcing = teacher_forcing
        self.dropout = nn.Dropout(dropout)
        self.bos_tag = self.vocab.stoi["xxbos"]

    def forward(self, features, captions_lengths, return_packed=False):
        """
        :param feature: features from encoder B, HXW, C
        :param captions_length: tuple containing captions and length
        :param return packed: tells whether to return packed sequence or not
        """
        # import pdb; pdb.set_trace();
        features = self.embed_features(features)
        captions, lengths = captions_lengths
        # import pdb; pdb.set_trace();
        embeddings = self.embed(captions)
        # we will not process xxeos token.
        packed = pack_padded_sequence(embeddings, lengths.cpu() - 1, batch_first=True)
        self.init_hidden(features)
        hx, cx = self.hx, self.cx
        outputs = []
        total_processed = 0
        next_words = (
            torch.tensor([self.bos_tag] * features.shape[0]).to(features.device).long()
        )
        for batch_size in packed[1]:
            features = features[:batch_size]  # batch_size, HXW, feature_space
            hx = hx[:batch_size]  # batch_size, hidden_dim
            cx = cx[:batch_size]  # batch_size, hidden_dim
            # context_vector: batch_size, attention_size
            # attention_map: batch_size, HXW, 1
            context_vector, attention_map = self.attention(features, hx)
            input_batch = packed[0][total_processed : total_processed + batch_size]
            # select input batch on based of teacher forcing prob.
            if self.teacher_forcing < random.random():
                input_batch = self.embed(next_words)
                input_batch = input_batch[:batch_size]

            # Adding attention image to input batch
            input_batch = torch.cat([input_batch, context_vector], dim=1)

            hx, cx = self.lstm(
                input_batch, (hx, cx)
            )  # batch_size, embedding_dim + attention_size
            current_op = self.out(self.dropout(hx))  # batch_size, vocab_size
            next_words = current_op.argmax(dim=-1)
            outputs.append(current_op)
            total_processed += batch_size

        return torch.cat(outputs), packed[1]

    def init_hidden(self, features):
        # mean does work of adaptive average pool
        self.hx = self.hidden_init(features.mean(1))
        self.cx = self.c_init(features.mean(1))

    def load_embedding(self, pretrained_embedding, vocab):
        # Initialize zeros vectors
        vectors = torch.zeros(len(vocab.itos), pretrained_embedding.get_dimension())
        # Load vectors from pretrained embeddings.
        for index, token in enumerate(vocab.itos):
            vectors[index] = torch.tensor(pretrained_embedding.get_word_vector(token))
        # Using pretrained vectors from fastext to initialize embeddings.
        layer = nn.Embedding.from_pretrained(vectors)
        return layer

    def decode_step(self, current_words, im, hidden):
        # This function is used in nucleus decoding for which
        # we have not given support
        with torch.no_grad():
            hx, cx = hidden
            embedding = self.embed(current_words)
            context_vector, attention_map = self.attention(im, hx[[0]])
            if embedding.shape[0] != context_vector.shape[0]:
                context_vector = context_vector[[0]].repeat(embedding.shape[0], 1)

            if embedding.shape[0] != hx.shape[0]:
                hx = hx[[0]].repeat(embedding.shape[0], 1)
                cx = cx[[0]].repeat(embedding.shape[0], 1)

            input_text = torch.cat([embedding, context_vector], dim=1)

            hx, cx = self.lstm(input_text, (hx, cx))
            out = self.out(hx)
        return out, (hx, cx), attention_map

    def sample_nucleus(self, im, p, max_len):
        # Nucleus decoding is not getting used.
        self.init_hidden(im)
        hidden = self.hx, self.cx
        predicted_sentence = []
        bos_tag = [self.vocab.stoi["xxbos"]]
        current_words = torch.tensor(bos_tag).to(im.device).long()
        for i in range(max_len):
            out, hidden, attention_map = self.decode_step(current_words, im, hidden)
            current_words = select_nucleus(out, p=p)
            print(current_words)
            predicted_sentence.append(current_words.item())
        return self.vocab.textify(predicted_sentence)

    def sample(self, features, beam_width, max_len):
        """
        takes in a batch of features
        """
        batch_size = features.shape[0]
        bos = self.vocab.stoi["xxbos"]
        eos = self.vocab.stoi["xxeos"]

        # decode step function it will be used in Beam Search.
        # Takes in images featurs, current input and hidden state
        def decode_step(self, im, current_words, hidden):
            with torch.no_grad():
                hx, cx = hidden["hx"], hidden["cx"]
                embedding = self.embed(current_words)
                context_vector, attention_map = self.attention(im, hx)
                input_text = torch.cat([embedding, context_vector], dim=1)
                hx, cx = self.lstm(input_text, (hx, cx))
                out = self.out(hx).log_softmax(-1)
            return out, {"hx": hx, "cx": cx, "attention_map": attention_map}

        # Initialize beam search with parameters.
        beam_search = BeamSearch(
            end_index=eos,
            max_steps=max_len,
            beam_size=beam_width,
            per_node_beam_size=beam_width // 2,
        )
        self.init_hidden(features)
        states = {"hx": self.hx, "cx": self.cx}
        start_predictions = torch.tensor([bos] * batch_size).to(features.device)
        # Initialize the function with features
        partial_decode_step = partial(decode_step, self, features)
        # Do beam search
        output = beam_search.search(start_predictions, states, partial_decode_step)

        captions = []
        # output[0] contains tokens of captions ranked
        # by probability.
        for caption in output[0]:
            caption = caption[0]
            captions.append(self.vocab.textify(caption[caption != eos]))

        return captions, (output[0], output[2])


class EncoderDecoder(nn.Module):
    def __init__(self, encoder, decoder):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder

    def forward(self, inputs, captions):
        features = self.encoder(inputs)
        return self.decoder(features, captions)

    def sample(self, input_batch, beam_width=8, max_len=25):
        self.eval()
        if type(input_batch) is tuple:
            input_batch, _ = input_batch

        features = self.encoder(input_batch)
        features = self.decoder.embed_features(features)
        captions, attention_maps = self.decoder.sample(
            features, beam_width=beam_width, max_len=max_len
        )
        return input_batch, captions, attention_maps


def loss_function_attention(inputs, captions, lengths):
    # skipping xxbos because that is going as input
    # and we need to predict next indexes onwards
    packed = pack_padded_sequence(captions[:, 1:], lengths.cpu() - 1, batch_first=True)
    return F.cross_entropy(inputs[0], packed[0])


def load_fasttext_embeddings(language="en"):
    embeddings_file = f"cc.{language}.300.bin"
    embeddings_path = os.path.join(Path.home(), ".cache", "embeddings")
    if not os.path.exists(embeddings_path):
        os.makedirs(embeddings_path)
    embeddings_file_path = os.path.join(embeddings_path, embeddings_file)
    # print(embeddings_file_path, os.path.exists(embeddings_file_path))
    if not os.path.exists(embeddings_file_path):
        fasttext.util.download_model(language, if_exists="ignore")
        shutil.move(embeddings_file, embeddings_path)
        os.remove(embeddings_file + ".gz")

    orig_stderr = sys.stderr
    temp_f = tempfile.TemporaryFile(mode="w")
    sys.stderr = temp_f
    ft = fasttext.load_model(embeddings_file_path)
    sys.stderr = orig_stderr
    return ft


def image_captioner_learner(
    data, backbone, attention=True, decoder_params=None, metrics=None, pretrained=True
):
    if attention:
        pretrained_embeddings = decoder_params.get("pretrained_embeddings", False)

        decoder_params = {
            "embed_size": decoder_params.get("embed_size", 100),
            "hidden_size": decoder_params.get("hidden_size", 100),
            "attention_size": decoder_params.get("attention_size", 100),
            "max_seq_length": decoder_params.get("max_seq_length", 20),
            "teacher_forcing": decoder_params.get("teacher_forcing", 1),
            "dropout": decoder_params.get("dropout", 0.1),
        }

        # Download pretrained embeddings if not already present.
        if pretrained_embeddings:
            ft = load_fasttext_embeddings(data.lang)
            if ft.get_dimension() != decoder_params["embed_size"]:
                decoder_params["embed_size"] = ft.get_dimension()
            decoder_params["pretrained_embeddings"] = ft

        decoder_params["vocab"] = data.vocab
        decoder_params["vocab_size"] = len(data.vocab.itos)
        # create ecoder using backbone
        encoder = EncoderAttention(backbone, pretrained=pretrained)
        # get channels from encoder which were computed during init.
        decoder_params["feature_size"] = encoder.feature_size
        # create LSTM decoder
        decoder = DecoderAttention(**decoder_params)
    else:
        raise NotImplementedError
    model = EncoderDecoder(encoder, decoder)

    # create learner using data.
    learn = Learner(
        data,
        model,
        loss_func=loss_function_attention,
        metrics=[accuracy, CorpusBLEU(len(data.vocab.itos))],
    )
    # Append callback
    learn.callbacks.append(ModifyInputCallback(learn))
    # learn.callbacks.append(ReduceTeacherForcing(learn))
    # split model for discriminative learning rate at embedding layer.
    split_layer_groups(learn)
    # return learner
    return learn


def split_layer_groups(learn):
    layers = flatten_model(learn.model)
    for idx, l in enumerate(layers):
        if "embed" in layers:
            break
    learn.layer_groups = split_model_idx(learn.model, [idx])


class ModifyInputCallback(LearnerCallback):
    def on_batch_begin(self, last_input, last_target, **kwargs):
        """
        Also include target in the input while training the model,
        because we do teacher forcing for the first 5 epochs.
        """
        return {"last_input": (last_input, last_target), "last_target": last_target}


class ReduceTeacherForcing(LearnerCallback):
    def on_epoch_end(self, **kwargs):
        """
        Reduce teacher forcing by a factor of 0.7 after each epoch.
        """
        if kwargs["epoch"] > 5:
            # 0.7 factor is experimental, can be improved based
            # on experimentation or defacto standards.
            # It will remain 1 for the first 5 epochs.
            self.learn.model.decoder.teacher_forcing *= 0.99


def accuracy(inputs, captions, lengths):
    # skipping xxbos because that is going as input
    # and we need to predict next indexes onwards
    packed = pack_padded_sequence(captions[:, 1:], lengths.cpu() - 1, batch_first=True)
    return (inputs[0].argmax(dim=-1) == packed[0]).float().mean()


class NGram:
    def __init__(self, ngram, max_n=5000):
        self.ngram, self.max_n = ngram, max_n

    def __eq__(self, other):
        if len(self.ngram) != len(other.ngram):
            return False
        return np.all(np.array(self.ngram) == np.array(other.ngram))

    def __hash__(self):
        return int(sum([o * self.max_n**i for i, o in enumerate(self.ngram)]))


def get_grams(x, n, max_n=5000):
    return (
        x
        if n == 1
        else [NGram(x[i : i + n], max_n=max_n) for i in range(len(x) - n + 1)]
    )


def get_correct_ngrams(pred, targ, n, max_n=5000):
    pred_grams, targ_grams = (
        get_grams(pred, n, max_n=max_n),
        get_grams(targ, n, max_n=max_n),
    )
    pred_cnt, targ_cnt = Counter(pred_grams), Counter(targ_grams)
    return sum([min(c, targ_cnt[g]) for g, c in pred_cnt.items()]), len(pred_grams)


def sentence_bleu(pred, targ, max_n=5000):
    corrects = [get_correct_ngrams(pred, targ, n, max_n=max_n) for n in range(1, 5)]
    n_precs = [c / t for c, t in corrects]
    len_penalty = exp(1 - len(targ) / len(pred)) if len(pred) < len(targ) else 1
    return len_penalty * ((n_precs[0] * n_precs[1] * n_precs[2] * n_precs[3]) ** 0.25)


def pad_tensor(tensor, batch_size):
    length = tensor.shape[0]
    n = batch_size - length
    return torch.cat([tensor, torch.zeros(n).to(tensor.device).long()], dim=0)[:, None]


def pack_into_batch(predictions, batches, batch_size):
    batch_of_captions = []
    total_processed = 0
    for l in batches.tolist():
        batch_of_captions.append(
            pad_tensor(predictions[total_processed : total_processed + l], batch_size)
        )
        total_processed += l
    return torch.cat(batch_of_captions, dim=1)


def post_process(predictions, gt_captions, batches):
    """
    predictions: padded sequence.
    gt_captions: padded_sequence ground truth.
    """
    batch_size = gt_captions.shape[0]
    predictions = predictions.argmax(dim=-1)
    padded_batch = pack_into_batch(predictions, batches, batch_size).cpu().numpy()
    gt_captions = gt_captions[:, 1:-1].cpu().numpy()
    pred_captions = []
    gt_captions_list = []
    for n in range(padded_batch.shape[0]):
        pred_captions.append(np.trim_zeros(padded_batch[n], trim="b"))
        gt_captions_list.append(np.trim_zeros(gt_captions[n], trim="b"))
    return pred_captions, gt_captions_list


class CorpusBLEU(Callback):
    def __init__(self, vocab_sz):
        self.vocab_sz = vocab_sz
        self.name = "bleu"
        self.pred_len, self.targ_len, self.corrects, self.counts = (
            0,
            0,
            [0] * 4,
            [0] * 4,
        )

    def on_epoch_begin(self, **kwargs):
        self.pred_len, self.targ_len, self.corrects, self.counts = (
            0,
            0,
            [0] * 4,
            [0] * 4,
        )

    def on_batch_end(self, last_output, last_target, **kwargs):
        preds, targs = post_process(last_output[0], last_target[0], last_output[1])
        for pred, targ in zip(preds, targs):
            self.pred_len += len(pred)
            self.targ_len += len(targ)
            for i in range(4):
                c, t = get_correct_ngrams(pred, targ, i + 1, max_n=self.vocab_sz)
                self.corrects[i] += c
                self.counts[i] += t

    def on_epoch_end(self, last_metrics, **kwargs):
        precs = [c / (t + EPS) for c, t in zip(self.corrects, self.counts)]
        len_penalty = (
            exp(1 - self.targ_len / (self.pred_len + EPS))
            if self.pred_len < self.targ_len
            else 1
        )
        bleu = len_penalty * ((precs[0] * precs[1] * precs[2] * precs[3]) ** 0.25)
        return add_metrics(last_metrics, bleu)


def predict_image(
    self, image, visualize, beam_width=5, max_len=20, visualize_attention=False
):
    import skimage
    import math

    # if PIL image/ ndarray convert to torch.tensor
    image = open_image(image)
    # normalize and resize
    image = image.resize(self._data.chip_size)
    image_tensor = normalize(image.px, *self._data.norm_stats)
    image_tensor = image_tensor[None].to(self._device)  # batch size 1
    # get captions
    _, caption, att_vars = self.learn.model.sample(
        image_tensor, beam_width=beam_width, max_len=max_len
    )
    if visualize:
        fig, ax = plt.subplots()
        show_image_and_text(ax, image.px, caption[0], False)

    if visualize_attention:
        cap = att_vars[0][0, 0]
        length = len(cap[cap != 3])
        b = 0  # batch index is zero
        nrows = math.ceil(length / 5)
        fig, axs = plt.subplots(nrows, 5, figsize=((20.0, 10.0)), squeeze=False)
        axs = axs.flatten()
        for k, atts in enumerate(att_vars[1]):
            if cap[k] != 3:
                ax = axs[k]
                attn_map = atts.view(1, beam_width, 7, 7)[b][0].cpu().numpy()
                image = skimage.transform.resize(attn_map, [224, 224])
                image = np.flip(image, 0)
                image = np.flip(image, 1)
                string = self._data.vocab.itos[cap[k].item()]
                ax.text(
                    0,
                    1,
                    "%s" % (string),
                    color="black",
                    backgroundcolor="white",
                    fontsize=12,
                )
                ax.imshow(image, cmap="gray")
            else:
                break

        for z in range(k, len(axs)):
            fig.delaxes(axs[z])

    return caption[0]


def get_bleu(self, data, beam_width=5, max_len=20):
    """
    Computes BLEU score considering after beam search.
    """
    captions, captions_gts = [], []

    # iterate over the validation set and
    # predictions and ground truth captions
    for x_in, current_caption_gt in data.valid_dl:
        captions_gts.extend(
            [cap.tolist()[: length.item()] for cap, length in zip(*current_caption_gt)]
        )
        _, current_captions, _ = self.learn.model.sample(
            x_in, beam_width=beam_width, max_len=max_len
        )
        captions.extend(current_captions)

    # compute counting of 1,2,3 and 4 grams
    pred_len, targ_len, corrects, counts = 0, 0, [0] * 4, [0] * 4
    for pred, targ in zip(captions, captions_gts):
        targ = targ[1:-1]
        pred = data.vocab.numericalize(data.tokenizer._process_all_1([pred])[0])
        pred_len += len(pred)
        targ_len += len(targ)
        for i in range(4):
            c, t = get_correct_ngrams(pred, targ, i + 1, max_n=len(data.vocab.itos))
            corrects[i] += c
            counts[i] += t

    # compute precision of all type of bleu
    n_precs = [c / (t + EPS) for c, t in zip(corrects, counts)]
    precs = n_precs

    # compute overall bleu as https://www.aclweb.org/anthology/P02-1040.pdf
    len_penalty = exp(1 - targ_len / (pred_len + EPS)) if pred_len < targ_len else 1
    bleu = len_penalty * ((precs[0] * precs[1] * precs[2] * precs[3]) ** 0.25)

    BLEU = {
        "bleu-1": n_precs[0],
        "bleu-2": n_precs[1],
        "bleu-3": n_precs[2],
        "bleu-4": n_precs[3],
        "BLEU": bleu,
    }

    return BLEU



import traceback
import warnings
from .._data import _raise_fastai_import_error
from ._inference_only_models import InferenceOnlyModel
from functools import partial


HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from fastprogress.fastprogress import progress_bar

    try:
        # For version 3.3.0
        from transformers.modeling_auto import MODEL_FOR_QUESTION_ANSWERING_MAPPING
    except ModuleNotFoundError as e:
        # For version 4.5.1
        from transformers.models.auto.modeling_auto import (
            MODEL_FOR_QUESTION_ANSWERING_MAPPING,
        )
    EXPECTED_MODEL_TYPES = [
        x.__name__.replace("Config", "")
        for x in MODEL_FOR_QUESTION_ANSWERING_MAPPING.keys()
    ]
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []


class QuestionAnswering(InferenceOnlyModel):
    """
    Creates a :class:`~arcgis.learn.text.QuestionAnswering` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify the HuggingFace
                            transformer model name which will be used to
                            extract the answers from a given passage/context.

                            To learn more about the available models for
                            question-answering task, kindly visit:-
                            https://huggingface.co/models?pipeline_tag=question-answering
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    pretrained_path         Option str. Path to a directory, where pretrained
                            model files are saved.
                            If pretrained_path is provided, the model is
                            loaded from that path on the local disk.
    ---------------------   -------------------------------------------
    working_dir             Option str. Path to a directory on local filesystem.
                            If directory is not present, it will be created.
                            This directory is used as the location to save the
                            model.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.QuestionAnswering` Object
    """

    #: supported transformer architectures
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None, **kwargs):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)
        super().__init__(backbone=backbone, task="question-answering", **kwargs)

    def get_answer(
        self,
        text_or_list,
        context,
        show_progress=True,
        explain=False,
        explain_start_word=True,
        explain_index=None,
        **kwargs,
    ):
        """
        Find answers for the asked questions from the given passage/context

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. Questions or a list
                                of questions one wishes to seek an answer for.
        ---------------------   -------------------------------------------
        context                 Required string. The context associated with
                                the question(s) which contains the answers.
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
        ---------------------   -------------------------------------------
        explain                 optional Bool. If set to True, will generate
                                a shap based explanation
        ---------------------   -------------------------------------------
        explain_start_word      optional Bool.
                                E.g. Context: Point cloud datasets are typically
                                collected using Lidar sensors (
                                light detection and ranging )
                                Question: "How is Point cloud dataset collected?"
                                Answer: Lidar Sensors

                                If set to True, will generate
                                a shap based explanation for start word. if set
                                to False, will generate explanation for last word
                                of the answer.

                                In the above example, if the value of `explain_start_word`
                                is `True`, it will generate the importance of different context
                                words that leads to selection of "Lidar" as a starting word
                                of the span. If `explain_start_word` is set to `False`
                                then it will generate explanation for the word `sensors`
        ---------------------   -------------------------------------------
        explain_index           optional List. Index of the question for which answer
                                needs to be generated
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        num_answers             Optional integer. The number of answers to
                                return. The answers will be chosen by order
                                of likelihood.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        max_answer_length       Optional integer. The maximum length of the
                                predicted answers.
                                Default value is set to 15.
        ---------------------   -------------------------------------------
        max_question_length     Optional integer. The maximum length of the
                                question after tokenization. Questions will be
                                truncated if needed.
                                Default value is set to 64.
        ---------------------   -------------------------------------------
        impossible_answer       Optional bool. Whether or not we accept impossible
                                as an answer.
                                Default value is set to False
        =====================   ===========================================

        :return: a list or a list of list containing the answer(s) for the input question(s)
        """
        results, kwargs_dict = [], {}

        kwargs_dict["topk"] = kwargs.get("num_answers", 1)
        kwargs_dict["max_answer_len"] = kwargs.get("max_answer_length", 15)
        kwargs_dict["max_question_len"] = kwargs.get("max_question_length", 64)
        kwargs_dict["handle_impossible_answer"] = kwargs.get("impossible_answer", False)

        if not isinstance(text_or_list, (list, tuple)):
            text_or_list = [text_or_list]

        for i in progress_bar(range(len(text_or_list)), display=show_progress):
            results.append(
                self.model(question=text_or_list[i], context=context, **kwargs_dict)
            )
        try:
            if explain:
                temp_text_or_list = []
                if explain_index is not None:
                    if isinstance(explain_index, list):
                        # validate the index
                        temp_index = []
                        invalid_index = []
                        for index in explain_index:
                            if isinstance(index, int):
                                temp_index.append(index)
                            else:
                                invalid_index.append(index)

                        if invalid_index:
                            with warnings.catch_warnings():
                                warnings.simplefilter("always", UserWarning)
                                warnings.warn(
                                    f"Index {invalid_index} are not valid. Indices/index must be integer. Ignoring "
                                    f"{invalid_index} for processing."
                                )

                        if temp_index:
                            for i in temp_index:
                                if i < len(text_or_list):
                                    temp_text_or_list.append(text_or_list[i])
                                else:
                                    with warnings.catch_warnings():
                                        warnings.simplefilter("always", UserWarning)
                                        warnings.warn(
                                            f"Value of index {i} should be less than/equal to {len(text_or_list) - 1}."
                                        )
                        else:
                            with warnings.catch_warnings():
                                warnings.simplefilter("always", UserWarning)
                                warnings.warn(
                                    f"No valid indices were supplied. Please change your input to list of integers"
                                )

                    elif isinstance(explain_index, int):
                        if explain_index < len(text_or_list):
                            temp_text_or_list = [text_or_list[explain_index]]
                        else:
                            with warnings.catch_warnings():
                                warnings.simplefilter("always", UserWarning)
                                warnings.warn(
                                    f"Value of index {explain_index} should be less than/equal to {len(text_or_list) - 1}."
                                )
                else:
                    temp_text_or_list = text_or_list

                if len(temp_text_or_list) > 0:
                    if explain_start_word:
                        self._explain(temp_text_or_list, context, True)
                    else:
                        self._explain(temp_text_or_list, context, False)
        except:
            with warnings.catch_warnings():
                warnings.simplefilter("always", UserWarning)
                warnings.warn(
                    f"SHAP workflow has encountered an error. Failed to generate an explanation."
                )

        return self._process_result(results, text_or_list)

    @staticmethod
    def _process_result(result_list, question_list):
        processed_results = []
        for result, question in zip(result_list, question_list):
            if isinstance(result, dict):
                tmp_dict = {
                    "question": question,
                    "answer": result["answer"],
                    "score": result["score"],
                }
                processed_results.append(tmp_dict)
            elif isinstance(result, list):
                item_list = [
                    {
                        "question": question,
                        "answer": item["answer"],
                        "score": item["score"],
                    }
                    for item in result
                ]
                processed_results.append(item_list)

        return processed_results

    def _logit_wrapper(self, part_start, questions):
        outs = []
        for q in questions:
            question, context = q.split("[SEP]")
            d = self.model.tokenizer(question, context, truncation="only_second")
            out = self.model.model.forward(
                **{k: torch.tensor(d[k]).reshape(1, -1).to(self._device) for k in d}
            )
            logits = out.start_logits if part_start else out.end_logits
            outs.append(logits.reshape(-1).detach().cpu().numpy())
        return outs

    def _output_token_decode(self, inputs):
        question, context = inputs.split("[SEP]")
        d = self.model.tokenizer(question, context, truncation="only_second")
        return [self.model.tokenizer.decode([id]) for id in d["input_ids"]]

    def _explain(self, questions, context, explain_start_token=True):
        IS_SHAP = True
        try:
            import shap
        except:
            IS_SHAP = False

        if IS_SHAP:
            logit_start = partial(self._logit_wrapper, True)
            logit_end = partial(self._logit_wrapper, False)
            notify = []
            for i in questions:
                try:
                    val = i + "[SEP]" + context
                    if explain_start_token:
                        explainer_start = shap.Explainer(
                            logit_start,
                            self.model.tokenizer,
                            output_names=self._output_token_decode(val),
                        )
                    else:
                        explainer_start = shap.Explainer(
                            logit_end,
                            self.model.tokenizer,
                            output_names=self._output_token_decode(
                                i + "[SEP]" + context
                            ),
                        )
                    shap_values_start = explainer_start([val])
                    shap.plots.text(shap_values_start)
                except ValueError as err:
                    notify.append(
                        f"SHAP based explanation failed for question {i} due to internal error."
                    )

            if len(notify):
                for i in notify:
                    print(i)
        else:
            warnings.warn("SHAP is not installed.")



from __future__ import annotations
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS
from typing import Dict, Any, Optional, List


class UpgradeManager(_BaseKube):
    _url = None
    _gis = None
    _con = None
    _properties = None

    def __init__(self, url: str, gis: GIS):
        self._gis = gis
        self._con = gis._con
        self._url = url

    @property
    def version(self) -> Dict[str, Any]:
        """
        Returns the current version for a deployment. When an patch or
        release is installed, this resource will update to reflect the
        information included in the update's version object as well as
        the job messages recorded during the upgrade process.
        """
        url = f"{self._url}/currentVersion"
        params = {
            "f": "json",
        }
        return self._con.get(url, params)

    @property
    def history(self) -> Dict[str, Any]:
        """
        Returns the transaction history for all upgrade and rollback jobs.

        :return: Dict[str, Any]
        """
        url = f"{self._url}/history"
        params = {
            "f": "json",
        }
        return self._con.get(url, params)

    @property
    def installed_updates(self) -> List[Dict[str, Any]]:
        """
        Returns a cumulative list of patches and releases that are installed in the deployment

        :return: List[Dict[str, Any]]
        """
        url = f"{self._url}/installed"
        params = {
            "f": "json",
        }
        return self._con.get(url, params).get("updates", [])

    @property
    def rollback_options(self) -> List[Dict[str, Any]]:
        """
        Returns a list of possible rollback options for the site, depending
        on the patch that is installed. The ID for the specific rollback
        version is passed as input for the `rollback` operation.

        :return: List[Dict[str, Any]]
        """
        url = f"{self._url}/checkRollback"
        params = {
            "f": "json",
        }
        return self._con.post(url, params).get("updates", [])

    def rollback(
        self,
        version: Dict[str, Any],
        settings: Optional[Dict[str, str]] = None,
    ) -> Dict[str, Any]:
        """
        This operation uninstalls a patch, removing the updates and fixes
        that had been applied to specific containers, and restoring the
        deployment to a previous, user-specified version of the software.
        The rollback operation cannot be performed for release-based
        updates.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        version                Required Dict[str, str]. The version of the deployment the operation
                               will rollback towards. This value can be retrieved from the
                               `rollback_options`.
        ------------------     --------------------------------------------------------------------
        settings               Optional Dict[str, str]. A configuration for patch settings.
                               This is only available at 10.9.1+.
        ==================     ====================================================================

        :return: Dict[str, Any]


        """
        url = f"{self._url}/rollback"
        params = {
            "f": "json",
            "versionManifest": version,
            "rollbackSettings": settings or {},
        }
        return self._con.post(url, params)

    # ---------------------------------------------------------------------
    def upgrades(
        self,
        version_manifest: str,
        settings: dict,
        manifest_file: str | None = None,
    ):
        """
        The upgrade operation upgrades, through either a patch or a
        release, an ArcGIS Enterprise on Kubernetes deployment to the
        current version.

        Before performing an upgrade, the unique ID associated with the
        patch or release must be retrieved from the version manifest using
        the available operation. The version manifest is a JSON array of
        version objects that contain update-specific information, including
        a JSON array of container objects that specify affected containers
        and include their name, checksum, and image values.

        Once the ID has been retrieved, you must also retrieve the required
        upgrade settings that will be passed through as part of the upgrade
        operation. Some settings will require user input before they can be
        used during an upgrade. For more information about current upgrade
        settings, see the Upgrade settings section below.

        Once the upgrade job request has been submitted, the deployment
        will either install a new patch on the base version or upgrade the
        entire deployment to the latest release. While the job is running,
        the upgrades resource will return detailed, real-time job messages
        and status information. The upgrades resource's child operations
        and resources will remain inaccessible for the duration of the
        upgrade.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        version_manifest       Optional Dict[str, str]. The unique ID associated with a patch or
                               release. You can get the version manifest ID for a patch or release
                               from the JSON view of the available operation.
        ------------------     --------------------------------------------------------------------
        settings               Optional Dict[str, str]. A JSON object containing details for
                               release upgrade settings. These settings, retrieved from the
                               getUpgradeSettings operation, must be included in the request for
                               the upgrade to be successful. Currently, the object supports the
                               following three upgrade settings: updateToLatestPatch,
                               licenseUpload, and volumesConfig. These settings are applicable to
                               ArcGIS Enterprise on Kubernetes versions 11.0 and later.
        ------------------     --------------------------------------------------------------------
        manifest_file          Optional String. The file containing the version manifest.
        ==================     ====================================================================

        :return: Dict[str, Any]

        """
        url = f"{self._url}/upgrade"
        params = {
            "f": "json",
            "versionManifest": version_manifest,
            "upgradeSettings": settings,
        }
        files = {"licenseUpload": manifest_file}
        if manifest_file:
            return self._con.post(url, params, files=files)
        else:
            return self._con.post(url, params)

    # ---------------------------------------------------------------------
    def import_manifest(self, manifest: str) -> dict:
        """
        The importManifest operation allows organization administrators to
        import the version manifest into a disconnected environment that
        can be used to discover available updates and releases and upgrade
        an ArcGIS Enterprise on Kubernetes deployment. The version manifest
        must be downloaded from My Esri, which requires an initial internet
        connection.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        manifest               Required String. The file containing the version manifest (.dat
                               file), used to discover available updates or releases for an ArcGIS
                               Enterprise on Kubernetes deployment.
        ==================     ====================================================================

        :returns: dict
        """
        url = f"{self._url}/importManifest"
        params = {
            "f": "json",
        }
        files = {"manifestFile": manifest}
        return self._con.post_multipart(url, params, files=files)

    # ---------------------------------------------------------------------

    def upgrade_settings(self, upgrade_id: str) -> dict:
        """
        The getUpgradeSettings operation returns the required upgrade
        settings, and their expected formats, needed for a specific
        release, applicable to ArcGIS Enterprise on Kubernetes versions
        11.0 and later. These settings must be passed through as values for
        the upgradeSettings parameter to successfully upgrade an ArcGIS
        Enterprise on Kubernetes deployment. Some upgrade settings may
        require their value property to be modified before being submitted
        as part of the upgrade operation. For example, when upgrading an
        ArcGIS Enterprise on Kubernetes deployment from version 10.9.1 to
        11.0, you will need to modify the value property for the
        licenseUpload JSON object.
        """
        url = f"{self._url}/getUpgradeSettings"
        params = {
            "f": "json",
            "upgradeId": upgrade_id,
        }
        return self._con.post(url, params)

    def available(self) -> Dict[str, List]:
        """
        This operation returns the version manifest, a cumulative list of
        release and patch versions that have been made available to an
        ArcGIS Enterprise organization.

        :return: Dict[str, List]
        """
        url = f"{self._url}/available"
        params = {
            "f": "json",
        }
        return self._con.post(url, params)



# MIT License

# Copyright (c) 2019 Hengshuang Zhao

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Based on https://github.com/hszhao/semseg

import torch
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
import torch

import math
from fastai.callbacks.hooks import hook_outputs, hook_output
from fastai.callbacks.hooks import model_sizes
from fastai.vision import flatten_model
from fastai.vision.models import unet
from fastai.basic_train import Learner
from fastai.vision import to_device
from ._PointRend import PointRendSemSegHead
from ._arcgis_model import _set_ddp_multigpu, _isnotebook
from ._timm_utils import get_backbone
from ._deeplab_utils import (
    get_dilation_index,
    get_hooks,
    get_last_module,
    add_dilation,
)


def initialize_weights(*models):
    for model in models:
        for module in model.modules():
            if isinstance(module, nn.Conv2d) or isinstance(module, nn.Linear):
                nn.init.kaiming_normal_(module.weight)
                if module.bias is not None:
                    module.bias.data.zero_()
            elif isinstance(module, nn.BatchNorm2d):
                module.weight.data.fill_(1)
                module.bias.data.zero_()


class _PyramidPoolingModule(nn.Module):
    """
    Creates the pyramid pooling module as in https://arxiv.org/abs/1612.01105
    Takes a feature map from the backbone and pools it at different scales
    according to the given pyramid sizes and upsamples it to original feature
    map size and concatenates it with the feature map.
    Code from https://github.com/hszhao/semseg.
    """

    def __init__(self, in_dim, reduction_dim, setting):
        super(_PyramidPoolingModule, self).__init__()
        self.features = []

        ## Creating modules for different pyramid sizes
        for s in setting:
            self.features.append(
                nn.Sequential(
                    nn.AdaptiveAvgPool2d(s),
                    nn.Conv2d(in_dim, reduction_dim, kernel_size=1, bias=False),
                    nn.BatchNorm2d(reduction_dim, momentum=0.95),
                    nn.ReLU(inplace=True),
                )
            )
        self.features = nn.ModuleList(self.features)

    def forward(self, x):
        x_size = x.size()
        out = [x]
        for f in self.features:
            ## Pass through the module which reduces its spatial size and then upsamples it.
            out.append(
                F.interpolate(f(x), x_size[2:], mode="bilinear", align_corners=True)
            )
        out = torch.cat(out, 1)
        return out


def _pspnet_unet(
    num_classes, backbone_fn, chip_size=224, pyramid_sizes=(1, 2, 3, 6), pretrained=True
):
    """
    Function which returns PPM module attached to backbone which is then used to form the Unet.
    """
    backbone = get_backbone(backbone_fn, pretrained)
    backbone_name = backbone_fn.__name__
    modify_dilation_index, _ = get_dilation_index(backbone_name)
    hookable_modules = get_last_module(backbone)
    add_dilation(backbone_fn, hookable_modules, modify_dilation_index)

    ## returns the size of various activations
    feature_sizes = model_sizes(backbone, size=(chip_size, chip_size))

    ## Get number of channels in the last layer
    num_channels = feature_sizes[-1][1]

    penultimate_channels = num_channels / len(pyramid_sizes)
    ppm = _PyramidPoolingModule(num_channels, int(penultimate_channels), pyramid_sizes)

    in_final = int(penultimate_channels) * len(pyramid_sizes) + num_channels

    # Reduce channel size after pyramid pooling module to avoid CUDA OOM error.
    final_conv = nn.Conv2d(
        in_channels=in_final, out_channels=512, kernel_size=3, padding=1
    )

    ## To make Dynamic Unet work as it expects a backbone which can be indexed.
    if "densenet" in backbone_name or "vgg" in backbone_name:
        backbone = backbone[0]
    layers = [*backbone, ppm, final_conv]
    return nn.Sequential(*layers)


class AuxPSUnet(nn.Module):
    """
    Adds auxillary loss to PSUnet.
    """

    def __init__(self, model, chip_size, num_classes):
        super(AuxPSUnet, self).__init__()
        self.model = model

        for idx, i in enumerate(flatten_model(self.model)):
            if hasattr(i, "dilation"):
                dilation = i.dilation
                dilation = dilation[0] if isinstance(dilation, tuple) else dilation
                if dilation > 1:
                    break

        self.hook = hook_output(flatten_model(model)[idx - 1])

        ## returns the size of various activations
        model_sizes(self.model, size=(chip_size, chip_size))

        ## Geting the stored parameters inside of the hook
        aux_in_channels = self.hook.stored.shape[1]
        del self.hook.stored
        self.aux_logits = nn.Conv2d(aux_in_channels, num_classes, kernel_size=1)

    def forward(self, x):
        out = self.model(x)
        out = F.interpolate(out, x.shape[2:], mode="bilinear", align_corners=True)
        if self.training:
            aux_l = self.aux_logits(self.hook.stored)
            ## Remove hook to free up memory
            self.hook.remove()
            return (
                out,
                F.interpolate(aux_l, x.shape[2:], mode="bilinear", align_corners=True),
            )
        else:
            return out


def _add_auxillary_branch_to_psunet(model, chip_size, num_classes):
    return AuxPSUnet(model, chip_size, num_classes)


class PSPUnet(nn.Module):
    """
    Keep model output and input size same.
    """

    def __init__(self, model):
        super(PSPUnet, self).__init__()
        self.model = model

    def forward(self, x):
        out = self.model(x)
        out = F.interpolate(out, x.shape[2:], mode="bilinear", align_corners=True)
        return out


class PSPNet(nn.Module):
    """
    Vanilla PSPNet
    """

    def __init__(
        self,
        num_classes,
        backbone_fn,
        chip_size=224,
        pyramid_sizes=(1, 2, 3, 6),
        pretrained=True,
        pointrend=False,
        keep_dilation=False,
    ):
        super(PSPNet, self).__init__()
        self.pointrend = pointrend
        self.backbone = get_backbone(backbone_fn, pretrained)
        backbone_name = backbone_fn.__name__
        modify_dilation_index, self.vgg = get_dilation_index(
            backbone_name, pointrend, keep_dilation
        )
        hookable_modules = get_last_module(self.backbone)
        add_dilation(backbone_fn, hookable_modules, modify_dilation_index)
        hooks = get_hooks(backbone_name, hookable_modules)

        ## Hook at the index where we need to get the auxillary logits out along with Fine-grained features
        self.hook = hook_outputs(hooks)

        ## returns the size of various activations
        feature_sizes = model_sizes(self.backbone, size=(chip_size, chip_size))

        if not self.vgg:
            ## Geting the number of channel persent in stored activation inside of the hook
            aux_in_channels = self.hook[0].stored.shape[1]
            ## Get number of channels in the last layer
            num_channels = feature_sizes[-1][1]
        else:
            aux_in_channels = self.hook[-2].stored.shape[1]
            num_channels = self.hook[-1].stored.shape[1]

        penultimate_channels = num_channels / len(pyramid_sizes)
        self.ppm = _PyramidPoolingModule(
            num_channels, int(penultimate_channels), pyramid_sizes
        )

        self.final = nn.Sequential(
            ## To handle case when the length of pyramid_sizes is odd
            nn.Conv2d(
                int(penultimate_channels) * len(pyramid_sizes) + num_channels,
                math.ceil(penultimate_channels),
                kernel_size=3,
                padding=1,
                bias=False,
            ),
            nn.BatchNorm2d(math.ceil(penultimate_channels)),
            nn.ReLU(inplace=True),
            nn.Dropout(0.1),
            nn.Conv2d(math.ceil(penultimate_channels), num_classes, kernel_size=1),
        )

        self.aux_logits = nn.Conv2d(aux_in_channels, num_classes, kernel_size=1)

        if self.pointrend:
            if self.vgg:
                point_num_channels = (
                    self.hook[-3].stored.shape[1] + self.hook[-4].stored.shape[1]
                )
                stride = chip_size / self.hook[-1].stored.shape[2]
            else:
                point_num_channels = self.hook[1].stored.shape[1]
                stride = chip_size / feature_sizes[-1][2]

            subdivision_steps = math.ceil(math.log(stride, 2))
            self.pointrend_head = PointRendSemSegHead(
                num_classes,
                point_num_channels,
                train_num_points=(chip_size / stride) ** 2,
                subdivision_num_points=(chip_size / (stride / 2)) ** 2,
                subdivision_steps=subdivision_steps,
            )

        initialize_weights(self.aux_logits)
        initialize_weights(self.ppm, self.final)

    def forward(self, x):
        x_size = x.size()
        x = self.backbone(x)
        features = self.hook.stored

        if self.vgg:
            x = self.ppm(features[-1])
            x = self.final(x)
        else:
            x = self.ppm(x)
            x = self.final(x)

        if self.pointrend:
            if self.vgg:
                pointrend_out = self.pointrend_head(x, [features[-4], features[-3]])
            else:
                pointrend_out = self.pointrend_head(x, [features[1]])

        result = F.interpolate(x, x_size[2:], mode="bilinear", align_corners=True)

        if self.training:
            if self.vgg:
                x = self.aux_logits(features[-2])
            else:
                x = self.aux_logits(features[0])

            x = F.interpolate(x, x_size[2:], mode="bilinear", align_corners=True)

            if self.pointrend:
                return result, x, pointrend_out
            else:
                return result, x

        else:
            if self.pointrend:
                if pointrend_out.shape[-1] != x_size[-1]:
                    pointrend_out = F.interpolate(
                        pointrend_out, x_size[2:], mode="bilinear", align_corners=True
                    )
                return pointrend_out
            else:
                return result


class DummyDistributed:
    "Dummy class to create a Learner since learner is created from fuction not a class. It will be used in case of multigpu training."

    def __getitem__(self, item):
        return eval("self." + item)


def _pspnet_learner(
    data,
    backbone,
    chip_size=224,
    pyramid_sizes=(1, 2, 3, 6),
    pretrained=True,
    pointrend=False,
    keep_dilation=False,
    **kwargs
):
    "Build psp_net learner from `data` and `arch`."
    model = to_device(
        PSPNet(
            data.c,
            backbone,
            chip_size,
            pyramid_sizes,
            pretrained,
            pointrend,
            keep_dilation,
        ),
        data.device,
    )
    if not _isnotebook():
        distributed_prep = DummyDistributed()
        _set_ddp_multigpu(distributed_prep)
        if distributed_prep._multigpu_training:
            learn = Learner(data, model, **kwargs).to_distributed(
                distributed_prep._rank_distributed
            )
            learn._map_location_multi_gpu = {
                "cuda:%d" % 0: "cuda:%d" % distributed_prep._rank_distributed
            }
        else:
            learn = Learner(data, model, **kwargs)
    else:
        learn = Learner(data, model, **kwargs)
    return learn


def _pspnet_learner_with_unet(
    data,
    backbone,
    chip_size=224,
    pyramid_sizes=(1, 2, 3, 6),
    pretrained=True,
    unet_aux_loss=False,
    vggv2=True,
    **kwargs
):
    "Build psunet learner from `data` and `arch`."
    model = unet.DynamicUnet(
        encoder=_pspnet_unet(data.c, backbone, chip_size, pyramid_sizes, pretrained),
        n_classes=data.c,
        last_cross=False,
    )

    if unet_aux_loss:
        model = _add_auxillary_branch_to_psunet(model, chip_size, data.c)
    elif vggv2:
        model = PSPUnet(model)
    if not _isnotebook():
        distributed_prep = DummyDistributed()
        _set_ddp_multigpu(distributed_prep)
        if distributed_prep._multigpu_training:
            learn = Learner(data, model, **kwargs).to_distributed(
                distributed_prep._rank_distributed
            )
            learn._map_location_multi_gpu = {
                "cuda:%d" % 0: "cuda:%d" % distributed_prep._rank_distributed
            }
        else:
            learn = Learner(data, model, **kwargs)
    else:
        learn = Learner(data, model, **kwargs)
    return learn


def isin(target, keep_indices):
    # import pdb; pdb.set_trace();
    old_shape = target.shape
    mask = torch.cat(
        [(target.view(-1) == k)[:, None] for k in keep_indices], dim=1
    ).any(1)
    mask = mask.view(old_shape).contiguous()
    return mask


def accuracy(input, target, ignore_mapped_class=[]):
    """Computes per pixel accuracy."""
    if isinstance(input, tuple):  # while training
        input = input[0]
    if ignore_mapped_class == []:
        target = target.squeeze(1)
        return (input.argmax(dim=1) == target).float().mean()
    else:
        target = target.squeeze(1)
        _, total_classes, _, _ = input.shape
        keep_indices = [i for i in range(total_classes) if i not in ignore_mapped_class]
        for k in ignore_mapped_class:
            input[:, k] = input.min() - 1
        targ_mask = isin(target, keep_indices)
        return (input.argmax(dim=1)[targ_mask] == target[targ_mask]).float().mean()



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
COCO dataset which returns image_id for evaluation.

Mostly copy-paste from https://github.com/pytorch/vision/blob/13b35ff/references/detection/coco_utils.py
"""

import torch
import torch.utils.data
from typing import Optional, List
from torch import Tensor

# from torchvision.ops.misc import _NewEmptyTensorOp
from torchvision.ops.boxes import box_area


@torch.no_grad()
def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    if target.numel() == 0:
        return [torch.zeros([], device=output.device)]
    maxk = max(topk)
    batch_size = target.size(0)

    _, pred = output.topk(maxk, 1, True, True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].view(-1).float().sum(0)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def inverse_sigmoid(x, eps=1e-5):
    x = x.clamp(min=0, max=1)
    x1 = x.clamp(min=eps)
    x2 = (1 - x).clamp(min=eps)
    return torch.log(x1 / x2)


def box_iou(boxes1, boxes2):
    area1 = box_area(boxes1)
    area2 = box_area(boxes2)

    lt = torch.max(boxes1[:, None, :2], boxes2[:, :2])  # [N,M,2]
    rb = torch.min(boxes1[:, None, 2:], boxes2[:, 2:])  # [N,M,2]

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    inter = wh[:, :, 0] * wh[:, :, 1]  # [N,M]

    union = area1[:, None] + area2 - inter

    iou = inter / union
    return iou, union


def generalized_box_iou(boxes1, boxes2):
    """
    Generalized IoU from https://giou.stanford.edu/

    The boxes should be in [x0, y0, x1, y1] format

    Returns a [N, M] pairwise matrix, where N = len(boxes1)
    and M = len(boxes2)
    """
    # degenerate boxes gives inf / nan results
    # so do an early check
    # print(boxes1)
    # print((boxes1[:, 2:] >= boxes1[:, :2]).all())
    # assert (boxes1[:, 2:] >= boxes1[:, :2]).all()
    # assert (boxes2[:, 2:] >= boxes2[:, :2]).all()
    iou, union = box_iou(boxes1, boxes2)

    lt = torch.min(boxes1[:, None, :2], boxes2[:, :2])
    rb = torch.max(boxes1[:, None, 2:], boxes2[:, 2:])

    wh = (rb - lt).clamp(min=0)  # [N,M,2]
    area = wh[:, :, 0] * wh[:, :, 1]

    return iou - (area - union) / area


def box_cxcywh_to_xyxy(x):
    x_c, y_c, w, h = x.unbind(-1)
    b = [(x_c - 0.5 * w), (y_c - 0.5 * h), (x_c + 0.5 * w), (y_c + 0.5 * h)]
    return torch.stack(b, dim=-1)


def box_xyxy_to_cxcywh(x):
    x0, y0, x1, y1 = x.unbind(-1)
    b = [(x0 + x1) / 2, (y0 + y1) / 2, (x1 - x0), (y1 - y0)]
    return torch.stack(b, dim=-1)


def _max_by_axis(the_list):
    # type: (List[List[int]]) -> List[int]
    maxes = the_list[0]
    for sublist in the_list[1:]:
        for index, item in enumerate(sublist):
            maxes[index] = max(maxes[index], item)
    return maxes


def nested_tensor_from_tensor_list(tensor_list: List[Tensor]):
    # TODO make this more general
    if tensor_list[0].ndim == 3:
        # TODO make it support different-sized images
        max_size = _max_by_axis([list(img.shape) for img in tensor_list])
        # min_size = tuple(min(s) for s in zip(*[img.shape for img in tensor_list]))
        batch_shape = [len(tensor_list)] + max_size
        b, c, h, w = batch_shape
        dtype = tensor_list[0].dtype
        device = tensor_list[0].device
        tensor = torch.zeros(batch_shape, dtype=dtype, device=device)
        mask = torch.ones((b, h, w), dtype=torch.bool, device=device)
        for img, pad_img, m in zip(tensor_list, tensor, mask):
            pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
            m[: img.shape[1], : img.shape[2]] = False
    else:
        raise ValueError("not supported")
    return NestedTensor(tensor, mask)


class NestedTensor(object):
    def __init__(self, tensors, mask: Optional[Tensor]):
        self.tensors = tensors
        self.mask = mask

    def to(self, device, non_blocking=False):
        cast_tensor = self.tensors.to(device, non_blocking=non_blocking)
        mask = self.mask
        if mask is not None:
            assert mask is not None
            cast_mask = mask.to(device, non_blocking=non_blocking)
        else:
            cast_mask = None
        return NestedTensor(cast_tensor, cast_mask)

    def record_stream(self, *args, **kwargs):
        self.tensors.record_stream(*args, **kwargs)
        if self.mask is not None:
            self.mask.record_stream(*args, **kwargs)

    def decompose(self):
        return self.tensors, self.mask

    def __repr__(self):
        return str(self.tensors)



from ._sharing import KbertnetesPy
from ._admin import KubernetesAdmin
from ._admin import WebAdaptorManager, ArchitectureManager
from ._admin import ExternalContentManager
from ._admin import DataStore, DataStores
from ._admin import Deployment, DeploymentManager, DeploymentProperty
from ._admin import Backup, BackupStore, BackupStoresManager, RecoveryManager
from ._admin import Job, JobManager
from ._admin import LanguageManager, LicenseManager, LogManager, Mode, Overview
from ._admin import KubeEnterpriseGroups, KubeEnterpriseUser
from ._admin import (
    KubeOrganization,
    KubeOrganizations,
    KubeOrgFederations,
    KubeOrgLicense,
    KubeOrgSecurity,
)
from ._admin import (
    KubeSecurity,
    KubeSecurityCert,
    KubeSecurityConfig,
    KubeSecurityIngress,
    KubeSecuritySAML,
)
from ._admin import KubeService, GPJobManager, ServicesManager
from ._admin import (
    Container,
    Indexer,
    Server,
    ServerDefaults,
    ServerManager,
    SystemManager,
)
from ._admin import TaskManager
from ._admin import UpgradeManager, Uploads, UsageStatistics



import os
import logging as _logging
from typing import Any, Optional
import arcgis
from arcgis.gis import GIS


_log = _logging.getLogger(__name__)

_use_async = False


def _convert_colorbrewer(cb_list: list, alpha: float = 1):
    """
    Helper to convert JS array derived from custom colorbrewer palette
    into format used by the renderers. See https://colorbrewer2.org/.
    Accepts HEX or RGB versions.

    Example:
    >>> palette_hex = ['#fee6ce','#fdae6b','#e6550d']
    >>> _convert_colorbrewer(palette_hex)
    [[254, 230, 206, 255], [253, 174, 107, 255], [230, 85, 13, 255]]

    >>> palette_rgb = ['rgb(254,230,206)','rgb(253,174,107)','rgb(230,85,13)']
    >>> _convert_colorbrewer(palette_rgb)
    [[254, 230, 206, 255], [253, 174, 107, 255], [230, 85, 13, 255]]

    """

    new_list = []

    # hex format
    if "#" in cb_list[0]:
        for color in cb_list:
            color = color.strip("#")
            length = len(color)
            new_color = [
                int(color[i : i + length // 3], 16)
                for i in range(0, length, length // 3)
            ]
            new_color.append(alpha * 255)
            new_list.append(new_color)

    # rgb format
    else:
        for color in cb_list:
            color = color[4:-1]
            tokens = color.split(",")
            new_color = [int(token) for token in tokens]
            new_color.append(alpha * 255)
            new_list.append(new_color)

    return new_list


def _format_colors(colors, alpha, cstep=None):
    """
    Helper to format any form of color input into something usable by the
    renderers. Strings will be tokenized into colormap names or converted
    to RGB numbers if they are HEX format. Exported colorbrewer lists will
    be converted to RGB lists, RGB lists will be left untouched. Palettable
    palettes will have their .colors attribute converted from tuples to lists.
    Anything else will default to "jet" colormap.

    Example:
    >>> hex_string = '#fee6ce, #fdae6b, #e6550d'
    >>> _format_colors(hex_string, 1)
    [[254, 230, 206, 255], [253, 174, 107, 255], [230, 85, 13, 255]]

    >>> cmap_string = "Spectral, jet, autumn"
    >>> _format_colors(cmap_string, 1)
    ['Spectral', 'jet', 'autumn']

    >>> import palettable.wesanderson as wa
    >>> wes_test = wa.Aquatic1_5
    >>> _format_colors(wes_test, 1)
    [[52, 36, 25, 255],
    [28, 64, 39, 255],
    [241, 201, 14, 255],
    [102, 88, 153, 255],
    [184, 147, 130, 255]]

    >>> premade_list = [[254, 230, 206, 255], [253, 174, 107, 255], [230, 85, 13, 255]]
    >>> _format_colors(premade_list, 1)
    [[254, 230, 206, 255], [253, 174, 107, 255], [230, 85, 13, 255]]

    """
    if colors is None:
        fmt_colors = "jet"
    if isinstance(colors, list):
        # single RGB + Alpha set
        if len(colors) == 4 and all([isinstance(i, int) for i in colors]):
            fmt_colors = [colors]

        # exported colorbrewer JSON array or list of colormaps
        elif all([isinstance(i, str) for i in colors]):
            if all(["#" in i for i in colors]) or all(["rgb(" in i for i in colors]):
                fmt_colors = _convert_colorbrewer(colors, alpha)
            else:
                fmt_colors = colors

        # list of RGB + Alpha set (already valid)
        elif all(
            [
                isinstance(color, list)
                and len(color) == 4
                and all([isinstance(i, int) for i in color])
                for color in colors
            ]
        ):
            fmt_colors = colors

        # doesn't fit any formats, default it
        else:
            fmt_colors = "jet"

    # palettable palette object
    elif hasattr(colors, "colors"):
        if cstep is not None:
            temp_map = colors.mpl_colormap
            color_tuple = temp_map(cstep, bytes=True)
            fmt_colors = [[int(i) for i in color_tuple]]
        else:
            fmt_colors = []
            for color in colors.colors:
                val_list = [val for val in color]
                val_list.append(alpha * 255)
                fmt_colors.append(val_list)

    # any form of string
    elif isinstance(colors, str):
        fmt_colors = colors.replace(" ", "")
        fmt_colors = fmt_colors.split(",")
        if all(["#" in i for i in fmt_colors]):
            fmt_colors = _convert_colorbrewer(fmt_colors, alpha)

    # default
    else:
        fmt_colors = ["jet"]
    return fmt_colors


def _get_list_value(index, array):
    """
    helper operation to loop a list of values regardless of the index value

    Example:
    >>> a = [111,222,333]
    >>> list_loop(15, a)
    111
    """
    if len(array) == 0:
        return None
    elif index >= 0 and index < len(array):
        return array[index]
    return array[index % len(array)]


def create_colormap(
    color_list: list,
    name: str = "temp_cmap",
    bins: int = 256,
):
    """
    The ``create_colormap`` function is a simple function that allows a
    user to make a matplotlib ``LinearSegmentedColormap`` object based on a
    custom assortment of colors. Accepts a list of RGB/RGB + alpha arrays.
    This function is great for visualizing potential color schemes to be used
    in mapping.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    color_list             Required list. List items must be of the format ``[R, G, B]``, where
                           R, G, and B represent values from 0 to 255, or any other format
                           where the first 3 values represent RGB values in bytes.
    ------------------     --------------------------------------------------------------------
    name                   Optional string. The name of the new colormap item. Defaults to
                           "temp_cmap".
    ------------------     --------------------------------------------------------------------
    bins                   Optional int. The number of bins (or uniquely retrievable colors)
                           the new colormap has. Defaults to 256.
    ==================     ====================================================================
    """
    from matplotlib.colors import LinearSegmentedColormap

    if len(color_list) < 2:
        raise ValueError("List must have multiple colors")

    mpl_colors = []
    for color in color_list:
        conv = (color[0] / 255, color[1] / 255, color[2] / 255)
        mpl_colors.append(conv)

    return LinearSegmentedColormap.from_list(name, mpl_colors, bins)


def export_map(
    web_map_as_json: Optional[dict] = None,
    format: str = """PDF""",
    layout_template: str = """MAP_ONLY""",
    gis: Optional[GIS] = None,
    **kwargs,
):
    """
    The ``export_map`` function takes the state of the :class:`~arcgis.mapping.WebMap` object (for example, included services, layer visibility
    settings, client-side graphics, and so forth) and returns either (a) a page layout or
    (b) a map without page surrounds of the specified area of interest in raster or vector format.
    The input for this function is a piece of text in JavaScript object notation (JSON) format describing the layers,
    graphics, and other settings in the web map. The JSON must be structured according to the WebMap specification
    in the ArcGIS Help.

    .. note::
        The ``export_map`` tool is shipped with ArcGIS Server to support web services for printing, including the
        preconfigured service named ``PrintingTools``.


    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    web_map_as_json        Web Map JSON along with export options. See the
                           `Export Web Map Specifications <https://developers.arcgis.com/rest/services-reference/exportwebmap-specification.htm>`_
                           for more information on structuring this JSON.
    ------------------     --------------------------------------------------------------------
    format                 Format (str). Optional parameter.  The format in which the map image
                           for printing will be delivered. The following strings are accepted:

                           For example:
                                PNG8

                           Choice list:
                                ['PDF', 'PNG32', 'PNG8', 'JPG', 'GIF', 'EPS', 'SVG', 'SVGZ']
    ------------------     --------------------------------------------------------------------
    layout_template        Layout Template (str). Optional parameter.  Either a name of a
                           template from the list or the keyword MAP_ONLY. When MAP_ONLY is chosen
                           or an empty string is passed in, the output map does not contain any
                           page layout surroundings.

                           For example - title, legends, scale bar, and so forth

                           Choice list:

                               | ['A3 Landscape', 'A3 Portrait',
                               | 'A4 Landscape', 'A4 Portrait', 'Letter ANSI A Landscape',
                               | 'Letter ANSI A Portrait', 'Tabloid ANSI B Landscape',
                               | 'Tabloid ANSI B Portrait', 'MAP_ONLY'].

                           You can get the layouts configured with your GIS by calling the :meth:`get_layout_templates <arcgis.mapping.get_layout_templates>` function
    ------------------     --------------------------------------------------------------------
    gis                    The :class:`~arcgis.gis.GIS` to use for printing. Optional
                           parameter. When not specified, the active GIS will be used.
    ==================     ====================================================================

    Returns:
        A dictionary with URL to download the output file.
    """

    from arcgis.geoprocessing import import_toolbox as _import_toolbox
    from arcgis.geoprocessing._tool import _camelCase_to_underscore
    from urllib import parse

    verbose = kwargs.pop("verbose", False)

    if gis is None:
        gis = arcgis.env.active_gis
    params = {
        "web_map_as_json": web_map_as_json,
        "format": format,
        "layout_template": layout_template,
        "gis": gis,
        "future": False,
    }
    params.update(kwargs)

    url = os.path.dirname(gis.properties.helperServices.printTask.url)
    tbx = _import_toolbox(url, gis=gis, verbose=verbose)
    basename = os.path.basename(gis.properties.helperServices.printTask.url)
    basename = _camelCase_to_underscore(parse.unquote_plus(parse.unquote(basename)))

    fn = getattr(tbx, basename)
    return fn(**params)


def get_layout_templates(gis: Optional[GIS] = None):
    """

    The ``get_layout_templates`` method returns the content of the :class:`~arcgis.gis.GIS` object's layout templates.

    .. note::
        The layout templates are formatted as a dictionary.

    .. note::
        See the
        `Get Layout Templates Info Task <https://utility.arcgisonline.com/arcgis/rest/directories/arcgisoutput/Utilities/PrintingTools_GPServer/Utilities_PrintingTools/GetLayoutTemplatesInfo.htm>`_
        for additional help on the ``get_layout_templates`` method.


    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS` object. The ``GIS`` on which ``get_layout_templates`` runs.

                           .. note::
                            If ``gis`` is not specified, the active GIS is used.

    ==================     ====================================================================

    Returns:
       ``output_json`` - The layout templates as Python dictionary


    """
    from arcgis.geoprocessing import DataFile
    from arcgis.geoprocessing._support import _execute_gp_tool

    param_db = {
        "output_json": (str, "Output JSON"),
    }
    return_values = [
        {"name": "output_json", "display_name": "Output JSON", "type": str},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = os.path.dirname(gis.properties.helperServices.printTask.url)
    kwargs = {"gis": gis}
    return _execute_gp_tool(
        gis,
        "Get Layout Templates Info Task",
        kwargs,
        param_db,
        return_values,
        _use_async,
        url,
    )


get_layout_templates.__annotations__ = {"return": str}



from __future__ import annotations
from arcgis.gis.kubernetes._admin._base import _BaseKube
from arcgis.gis import GIS
from typing import Dict, Any, Optional, List
import time
import concurrent.futures


def sleep_counter(start=1, mval=6):
    while True:
        if start < mval:
            start += 1
            yield start
        else:
            return mval


def _status(gis: GIS, url: str, params: dict[str, Any] | None = None) -> dict[str, Any]:
    """Checks the status of a URL"""
    if params is None:
        params = {"f": "json"}
    res = gis._con.get(res, params)
    i = 1
    while res.get("status", "completed").lower() != "executing":
        res = gis._con.get(res, params)
        i += 1
        if i > 5:
            i = 5
        time.sleep(i)
    return res


###########################################################################
class BackupStore(_BaseKube):
    """ """

    _con = None
    _gis = None
    _url = None

    def __init__(self, url: str, gis: GIS) -> None:
        super()
        self._url: str = url
        self._gis: GIS = gis
        self._con = gis._con

    def update(self, settings: dict[str, Any]) -> bool:
        """

        Update only supports setting the backup store as the default store for your deployment {"default": true}.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        settings               Required dict[str, Any]. A JSON object of backup store settings.
                               At 10.9.1, the only supported setting is the default property.
                               Setting the default property as true will mark the backup store as
                               the default store for your deployment.
        ==================     ====================================================================

        :returns: bool
        """
        url: str = f"{self._url}/update"
        params: dict[str, Any] = {"f": "json", "settings": settings}
        return self._gis._con.post(url, params).get("status", "failed") == "success"

    def validate(self) -> dict[str, Any]:
        """
        This operation ensures that the backup store is able to access the object store and is ready for backup operations to be performed.
        """
        url: str = f"{self._url}/validate"
        params: dict[str, Any] = {"f": "json"}
        return self._gis._con.post(url, params)

    # ---------------------------------------------------------------------
    def delete(self) -> bool:
        """Unregisters a backup store from the deploayment"""
        url = f"{self._url}/unregister"
        params = {"f": "json"}
        return self._con.post(url, params).get("status", "failed") == "success"


class BackupStoresManager:
    """
    Manages the backup stores with the deployments
    """

    _url = None
    _gis = None
    _properties = None

    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis

    @property
    def properties(self) -> dict[str, Any]:
        """
        returns the endpoint properties

        :returns: dict[str, Any]
        """
        return self._gis._con.get(self._url, {"f": "json"})

    def register(
        self,
        name: str,
        settings: dict[str, Any],
        default: bool,
        future: bool = False,
    ) -> dict[str, Any] | concurrent.futures.Future:
        """
        The register operation registers a backup store. The backup store is created and managed by the deployment.

        :returns: dict | concurrent.futures.Future
        """
        params = {
            "f": "json",
            "storeName": name,
            "settings": settings,
            "isDefault": default,
            "async": True,
        }
        url = f"{self._url}/register"
        res = self._gis._con.post(url, params)
        url = res.get("jobsUrl", None)
        executor = concurrent.futures.ThreadPoolExecutor(1)
        executor.shutdown(True)
        return executor.submit(_status, **{"gis": self._gis, "url": url})

    def list(self):
        """Returns a list of all the backup stores objects"""
        return [
            BackupStore(url=f"{self._url}/{bck['name']}", gis=self._gis)
            for bck in self.properties.get("backupStores", [])
        ]


###########################################################################
class Backup(_BaseKube):
    """ """

    _con = None
    _gis = None
    _url = None

    def __init__(self, url: str, gis: GIS) -> None:
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    def delete(self) -> bool:
        """
        Removes the backup from the system

        :return: Boolean. True if successful else False.
        """
        url = f"{self._url}/delete"
        params = {"f": "json"}
        return self._con.post(url, params).get("status") == "success"

    # ---------------------------------------------------------------------
    def validate(self) -> dict:
        """
        This operation ensures that the backup store is able to access the
        object store and is ready for backup operations to be performed.

        :returns: Dict
        """
        url = f"{self._url}/validate"
        params = {"f": "json"}
        return self._con.post(url, params)

    def restore(self, store_name: str, passcode: str) -> bool:
        """
        Restores the organization to the state it was in when the backup
        was created. Once completed, any existing content and data will be
        replaced with what was contained in the backup.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        store_name             Required String. The name of the store the backup was copied to.
        ------------------     --------------------------------------------------------------------
        passcode               Required String. The passcode used to encrypted the backup. This
                               passcode must be the same as the one used when creating the backup.
        ==================     ====================================================================

        :return: Boolean. True if successful else False.

        """
        url = f"{self._url}/restore"
        params = {
            "f": "json",
            "storeName": store_name,
            "passcode": passcode,
        }
        return self._con.post(url, params).get("status") == "success"


###########################################################################
class RecoveryManager(_BaseKube):
    """
    Allows an administrator the ability to manage disaster recovery settings
    """

    _con = None
    _gis = None
    _url = None

    def __init__(self, url: str, gis: GIS) -> None:
        super()
        self._url = url
        self._gis = gis
        self._con = gis._con

    def register(
        self,
        name: str,
        credentials: Dict[str, str],
        root: str,
        storage_config: Dict[str, str],
        is_default: Optional[bool] = False,
    ) -> BackupStore:
        """
        This method registers a backup store.


        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required String. The unique name of the backup store.
        ------------------     --------------------------------------------------------------------
        credentials            Required Dict[str, str]. The credentials of the configuration store.
        ------------------     --------------------------------------------------------------------
        root                   Required String. The root directory of the store.
        ------------------     --------------------------------------------------------------------
        storage_config         Required Dictionary. A dictionary describing the storage
                               configuration for the backup store.
        ------------------     --------------------------------------------------------------------
        is_default             Optional Bool. Determines if the store will be the default backup
                               store.  The default is `False`.
        ==================     ====================================================================

        :return: BackupStore

        """
        try:
            url = f"{self._url}/stores/register"
            params = {
                "f": "json",
                "storeName": name,
                "credentials": credentials,
                "rootDirectory": root,
                "storageConfig": storage_config,
                "isDefault": is_default,
            }
            res = self._con.post(url, params)
            url = f"{self._url}/stores/{res['name']}"
            return BackupStore(url=url, gis=self._gis)
        except Exception as e:
            raise e

    def backup(
        self,
        name: str,
        store_name: str,
        passcode: str,
        description: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Creates a backup that can be restored in the event of data loss,
        data corruption, or deployment failures. Backups are stored in a
        designated backup store.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required String. The unique name of the backup.
        ------------------     --------------------------------------------------------------------
        store_name             Required String. The name of the backup store to save in.
        ------------------     --------------------------------------------------------------------
        passcode               Required String. A passcode that will be used to encrypt the content
                               of the backup. When restoring a backup, this passcode must be passed
                               in. The passcode must be at least eight characters in length.
        ------------------     --------------------------------------------------------------------
        description            Optional String. A description of the backup.
        ==================     ====================================================================

        :return: Dict[str, Any]
        """
        params = {
            "f": "json",
            "name": name,
            "storeName": store_name,
            "passcode": passcode,
            "description": description or "",
        }
        url = f"{self._url}/backuprestore/backup"
        return self._con.post(url, params)

    @property
    def backup_status(self) -> Dict[str, Any]:
        """
        This resource returns the status of a current, or previously
        executed, backup.

        :return: Dict[str, Any]

        """
        try:
            url = f"{self._url}/backuprestore/status"
            params = {"f": "json"}
            return self._con.get(url, params).get("status", {})
        except:
            return {}

    @property
    def backupstores(self) -> BackupStoresManager:
        """
        Manages the backup stores registered with the deployment

        :returns: BackupStoresManager
        """
        url = f"{self._url}/stores"
        return BackupStoresManager(url, self._gis)

    @property
    def backups(self) -> List[Backup]:
        """
        Returns the backups that have been created of an organization.

        :return: List[Backup]

        """
        try:
            url = f"{self._url}/backuprestore/backups"
            params = {"f": "json"}
            res = self._con.get(url, params)
            backups = []
            for bckup in res.get("backups", []):
                name = bckup["backupName"]
                url = f"{self._url}/backuprestore/backups/{name}"
                backups.append(Backup(url=url, gis=self._gis))
            return backups
        except:
            return []

    @property
    def stores(self) -> List[BackupStore]:
        """
        Returns the backup stores that are registered with your deployment.

        :return: List[BackupStore]
        """
        url = f"{self._url}/stores"
        params = {"f": "json"}
        res = self._con.get(url, params)
        stores = []
        for bs in res.get("backupStores", []):
            name = bs["name"]
            url = f"{self._url}/stores/{name}"
            stores.append(BackupStore(url=url, gis=self._gis))
        return stores

    @property
    def settings(self) -> Dict[str, Any]:
        """
        Gets/Sets the currently configured disaster recovery settings.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        value                  Required Dict[str, Any]. Dictionary describing disaster recovery
                               settings.
        ==================     ====================================================================

        :return: Dict[str, Any]

        """
        url = f"{self._url}/settings"
        params = {"f": "json"}
        return self._con.get(url, params)

    # ----------------------------------------------------------------------
    @settings.setter
    def settings(self, value: Dict[str, Any]) -> None:
        """
        See main ``settings`` property docstring
        """
        url = f"{self._url}/settings/update"
        params = {
            "f": "json",
            "settings": value,
        }
        return self._con.post(url, params)
